/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Danesh Kamal
 */

/**
 * @module nmodule/alarm/rc/Alarm
 * @private
 */
define([
  'baja!',
  'baja!alarm:AlarmRecord,box:AlarmChannel,alarm:AlarmConsoleOptions',
  'Promise',
  'jquery',
  'underscore'
], function (baja,
             types,
             Promise,
             $,
             _) {
  'use strict';

  var Alarm = {};

  /**
   * Asynchronously retrieve alarms for a given dynamic time range
   *
   * @param {bql.DynamicTimeRange} timeRange - Time range within which to query for results
   * @param {Integer} [limit=-1] - Maximum number of alarms to retrieve
   * @param {Integer} [offset=-1] - Starting offset of result cursor
   * @param {bql.FilterSet} [filterSet] - Filter to apply to query
   * @param {Boolean} [sortDescending=false] - Retrieve results in descending order
   * @param {String} [columnName=timestamp] - columnName on which to order results
   * @param {Object} alarmClassDisplayNames - Map of alarm class names to display names
   * @param {baja.TimeZoneDatabase} timezoneDatabase - Local timezone database for resolving Time Zone ids.
   * @returns {Promise} Promise - Promise resolved with results object
   */
  Alarm.retrieveAlarms = function (timeRange, limit, offset, filterSet, sortDescending, columnName, alarmClassDisplayNames, timezoneDatabase) {
    return baja.Ord.make('alarm:').get()
      .then(function (space) {
        return space.queryAlarmDatabase(timeRange, limit, offset, filterSet, sortDescending, columnName);
      })
      .then(function (result) {
        return decodeAlarms(alarmClassDisplayNames, timezoneDatabase, result.r)
          .then(function (alarms) {
            return {
              alarms: alarms,
              count: result.c
            };
          });
      });
  };

  function decodeAlarms(alarmClassMap, tzDatabase, records) {
    return Promise.all(_.map(records, function (record) {
      return baja.bson.decodeAsync(record, baja.$serverDecodeContext)
        .then(function (decoded) {
          return updateRecord(alarmClassMap, tzDatabase, decoded);
        });
    }));
  }

  /**
   * Update the AlarmRecord with the correct baja.TimeZone and Alarm Class Display Name
   * as well as adding properties for source and ack state values.
   */
  function updateRecord(alarmClassDisplayNames, tzDatabase, rec) {
    rec.alarmData = _.mapObject(rec.getAlarmData().toObject(), function (val, key) {
        switch (key) {
          case 'TimeZone':
            return tzDatabase.getTimeZone(val.getId());
          default:
            return val instanceof baja.Simple ? val.encodeToString() : val;
        }
      });

    rec.setAlarmClass(alarmClassDisplayNames[rec.getAlarmClass()]);
    rec._sourceState = rec.getSourceState().getTag();
    rec._ackState = rec.getAckState().getTag();
    return rec;
  }

/////////////////////////////////////////////////////////////////////
//Alarm Database Maintenance
/////////////////////////////////////////////////////////////////////

  /**
   * Retrieve the number of alarms in the database for a given time range (null arg returns total alarms)
   * @returns {Promise} Promise resolved with the alarm count
   */
  Alarm.getAlarmCount = function (timeRange, filterSet) {
    return baja.Ord.make("alarm:").get().then(function (space) {
      return space.getRecordCount(timeRange, filterSet);
    });
  };

  /**
   * Asynchronously clear all alarms in the alarm database
   * @returns {Promise} Promise
   */
  Alarm.clearAllRecords = function (callbacks) {
    return baja.Ord.make('alarm:').get().then(function (space) {
      return space.clearAllRecords({
        ok: function (response) {
          callbacks.ok(response);
        },

        fail: function (err) {
          callbacks.fail(err);
        }
      });
    });
  };

  /**
   * Asynchronously clear alarms records that are older than the given
   * timestamp
   *
   * @param {baja.AbsTime} [beforeTime] Records with timestamps older than this timestamp will be cleared
   * @returns {Promise} Promise
   * @param callbacks
   */
  Alarm.clearOldRecords = function (beforeTime, callbacks) {
    return baja.Ord.make('alarm:').get().then(function (space) {
      return space.clearOldRecords(beforeTime, {
        ok: function (response) {
          callbacks.ok(response);
        },

        fail: function (err) {
          callbacks.fail(err);
        }
      });
    });
  };

  /**
   * Asynchronously clear the alarm records specified by the uuids
   *
   * @param {Array} uuids  Array of encoded baja.Uuid strings
   * @returns {Promise} Promise
   * @param callbacks
   */
  Alarm.clearRecords = function (uuids, callbacks) {
    return baja.Ord.make('alarm:').get().then(function (space) {
      return space.clearRecords(uuids, {
        ok: function (response) {
          callbacks.ok(response);
        },

        fail: function (err) {
          callbacks.fail(err);
        }
      });
    });
  };

  /**
   * Asynchronously add a note to an alarm record
   * @param alarm The alarm record to which to add the new note
   * @param note
   * @param callbacks
   * @returns {Promise} Promise
   */
  Alarm.addNote = function (alarm, note, callbacks) {
    return baja.Ord.make('alarm:').get().then(function (space) {
      return space.addNote(alarm, note, {
        ok: function (response) {
          callbacks.ok(response);
        },

        fail: function (err) {
          callbacks.fail(err);
        }
      });
    });
  };

  /**
   * Asynchronously retrieves a map of alarm class names to alarm class display names
   * @returns {Promise} Promise resolved with object map
   */
  Alarm.getDisplayNamesMap = function () {
    return baja.Ord.make('alarm:').get().then(function (space) {
      return space.getDisplayNamesMap();
    });
  };

  return Alarm;
});
