/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Defines {@link baja.HierarchyScheme}.
 * @module baja/ord/HierarchyScheme
 */
define([
  "bajaScript/sys",
  "bajaScript/baja/ord/Ord",
  "bajaScript/baja/ord/OrdQuery",
  "bajaScript/baja/ord/OrdScheme",
  "bajaScript/baja/ord/OrdTarget"
], function (baja, Ord, OrdQuery, OrdScheme, OrdTarget) {
  
  "use strict";
  
  var subclass = baja.subclass,
      callSuper = baja.callSuper,
      hierarchySpaceTypeSpec = "hierarchy:HierarchySpace",
      hierarchyLevelElemTypeSpec = "hierarchy:LevelElem",
      virtualComponentTypeSpec = "baja:VirtualComponent";

  /**
   * Before resolving any hierarchy: ORD, the hierarchy space must be resolved
   * and associated type specs must be loaded. In turn, this will lazily resolve
   * the BajaScript library for managing hierarchies from the hierarchy module
   * (HierarchySpace Type Extension).
   * @function
   */
  var ensureSpaceResolved = (function () {
    var prom;
    return function () {
      return prom || (prom = baja.Ord.make("hierarchy:")
        .resolve({ forceServerResolve: true })
        .then(function () {
          return baja.importTypes({
            typeSpecs: [
              hierarchySpaceTypeSpec,
              hierarchyLevelElemTypeSpec,
              virtualComponentTypeSpec
            ]
          });
        }));
    };
  }());

  /**
   * Hierarchy ORD Scheme.
   *
   * This ORD scheme is used for resolving Niagara 4 hierarchies. The hierarchy
   * module must be installed on the Station in order for this to successfully
   * resolve.
   *
   * @class
   * @alias baja.HierarchyScheme
   * @extends baja.OrdScheme
   * @private
   */
  var HierarchyScheme = function HierarchyScheme() {
    callSuper(HierarchyScheme, this, arguments);
  };

  subclass(HierarchyScheme, OrdScheme);

  /**
   * Default Handle ORD Scheme instance
   * @private
   * @type {baja.HierarchyScheme}
   */
  HierarchyScheme.DEFAULT = new HierarchyScheme();

  /**
   * Called when an ORD is resolved.
   *
   * @private
   *
   * @see baja.OrdScheme#resolve
   *
   * @param {module:baja/ord/OrdTarget} target  the current ORD Target.
   * @param {baja.OrdQuery} query  the ORD Query used in resolving the ORD.
   * @param {module:baja/ord/OrdQueryListCursor} cursor  the ORD Query List
   * cursor used for helping to asynchronously resolve the ORD.
   * @param {Object} options  options used for resolving an ORD.
   */
  HierarchyScheme.prototype.resolve = function (target, query, cursor, options) {
    ensureSpaceResolved()
      .then(function () {
        resolveHierarchy(target, query, cursor, options);
      })
      .catch(function (err) {
        return options.callback.fail(err);
      });
  };

  function resolveHierarchy(target, query, cursor, options) {

    function fail (err) {
      options.callback.fail(err);
    }

    // Create the hierarchy space node if it does not exist already
    if (!baja.nav.localhost.hierarchy) {
      baja.nav.localhost.hierarchy = baja.nav.localhost.$addChildNode(
        baja.$(hierarchySpaceTypeSpec));
    }

    var newTarget = new OrdTarget(target);

    if (!query.getBody()) {
      newTarget.object = baja.nav.localhost.hierarchy;
      cursor.resolveNext(newTarget, options);
    }
    else {
      // We used to short-circuit if the last item in the query body started with "station:|".  This
      // would, however, resolve a component even if the user did not have permission to view the
      // hierarchy listed at the beginning of the query body.  We did not want to add another
      // network call just to check the user's permissions on a particular hierarchy.  Also, we did
      // not want to start caching hierarchy space nav children at this time.  That would possibly
      // prevent a network call: if the hierarchy is in the nav children, go ahead and
      // short-circuit.  But, refreshing this nav child cache would have to be considered.

      // Resolve the ORD and force it to resolve on the Server.
      baja.Ord.make(query.toString())
        .resolve({forceServerResolve: true})
        .then(function (target) {
          newTarget.object = target.object;
          var newTargetType = newTarget.object.getType();

          if (!newTargetType.is(hierarchyLevelElemTypeSpec)) {
            // Hierarchy ord resolved to a component or virtual component and not a
            // hierarchy:LevelElem
            if (!newTarget.object.isMounted()) {
              var resolvePromise;

              if (newTargetType.is(virtualComponentTypeSpec)) {
                // Cannot resolve virtual components by handle because its handle is in the virtual
                // space.  Use virtual info to form full ord:
                // <gatewayOrd>|virtual:|<slot ord in remote station>.
                // virtualInfo.getGateway() returns station:|<gateway handleOrd>
                // virtualInfo.getSlotOrd() returns station:|<slot ord in remote station>
                var virtualInfo = newTarget.object.getNiagaraVirtualCompInfo();
                var virtualOrd = virtualInfo.getGateway() + "|virtual:|" +
                  virtualInfo.getSlotOrd().parse().get('slot');
                resolvePromise = Ord.make(virtualOrd).get();
              }
              else {
                // Resolve the component by handle to properly load it in the ComponentSpace.
                resolvePromise = baja.station.resolveByHandle({
                  "handle": newTarget.object.getHandle()
                });
              }

              resolvePromise
                .then(function (comp) {
                  newTarget.object = comp;
                  return newTarget.object.loadSlots();
                })
                .then(function () {
                  cursor.resolveNext(newTarget, options);
                })
                .catch(fail);
            }
            else {
              // Non-hierarchy:LevelElem is already mounted; see NCCB-25269
              cursor.resolveNext(newTarget, options);
            }
          }
          else {
            // hierarchy:LevelElem
            cursor.resolveNext(newTarget, options);
          }
        }, fail);
    }
  }

  /**
   * Return an ORD Query for the scheme.
   *
   * @returns {baja.OrdQuery}
   */
  HierarchyScheme.prototype.parse = function (schemeName, body) {
    return new OrdQuery({
      scheme: this,
      schemeName: schemeName,
      body: body,
      isHost: false,
      isSession: false
    });
  };
  
  return HierarchyScheme;
});
