/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * @module nmodule/webChart/rc/command/ExportWizardCommand
 * @private
 */
define([
    'baja!',
    'jquery',
    'Promise',
    'dialogs',
    'nmodule/webChart/rc/webChartUtil',
    'nmodule/webChart/rc/choiceUtil',
    'nmodule/webChart/rc/export/exportUtil',
    'nmodule/webChart/rc/command/DialogWizardCommand',
    'nmodule/webChart/rc/line/TitleLayer',
    'nmodule/webChart/rc/chartEvents',
    'nmodule/webChart/rc/dragUtil',
    'bajaux/commands/Command',
    'nmodule/webChart/rc/tab/Tab',
    'nmodule/webChart/rc/tab/TabbedEditor',
    'nmodule/webChart/rc/fe/SimplePropertySheet',
    'baja!' +
    'baja:NameMap'
  ], function(baja,
    $,
    Promise,
    dialogs,
    webChartUtil,
    choiceUtil,
    exportUtil,
    DialogWizardCommand,
    TitleLayer,
    chartEvents,
    dragUtil,
    Command,
    Tab,
    TabbedEditor,
    SimplePropertySheet) {
    "use strict";

    /**
     * ExportWizardCommand provides options for exporting the ChartWidget.
     *
     * @private
     * @class
     * @alias module:nmodule/webChart/rc/command/ExportWizardCommand
     * @extends bajaux.Command
     */
    var ExportWizardCommand = function ExportWizardCommand(widget) {
      var that = this;
      that.$widget = widget;

      //TODO: bajascript access to BIFile mime types
      this.$defaultMimeTypes = {
        csv: "text/csv;charset=utf-8",
        chart: "application/json;charset=utf-8"
      };
      DialogWizardCommand.apply(this, arguments);
    };

    var setEnumChoice = choiceUtil.setEnumChoice,
      setBooleanChoice = choiceUtil.setBooleanChoice,
      setDisplayNames = choiceUtil.setDisplayNames,
      getChoice = choiceUtil.getChoice,
      getOrdsFromDrag = dragUtil.getOrdsFromDrag;

    ExportWizardCommand.prototype = Object.create(DialogWizardCommand.prototype);
    ExportWizardCommand.prototype.constructor = ExportWizardCommand;

    /**
     * Override resolveTab to resolve to a tab for the content.
     * @return {Promise}
     */
    ExportWizardCommand.prototype.resolveTab = function() {
      var that = this,
          widget = that.$widget;
      return TitleLayer.resolveTitleInfo(widget)
        .then(function(titleInfo) {
          var actionInfo = new baja.Component(),
              isWb = window.niagara &&
                window.niagara.env &&
                window.niagara.env.type === "wb",
              isIe = window.navigator && window.navigator.userAgent && window.navigator.userAgent.indexOf("MSIE ") > 0,
              destinationOptions = ["station"];

          actionInfo.add({
            slot: 'fileName',
            value: titleInfo.title
          });

          if(!isWb && !isIe){
            destinationOptions.push("print");
            destinationOptions.unshift("download");
          }

          setEnumChoice(actionInfo, "destination", destinationOptions);
          setEnumChoice(actionInfo, "fileType", ["chart", "csv"]);
          setBooleanChoice(actionInfo, "statusColumn", false);
          setBooleanChoice(actionInfo, "viewOnExport");

          setEnumChoice(actionInfo, "ordType", ["absolute", "relative"]);

          // Get base folder / ord
          var baseOrd;
          if (typeof window.niagara !== 'undefined' &&
            window.niagara.env &&
            typeof window.niagara.env.getBaseOrd === 'function') {
            baseOrd = window.niagara.env.getBaseOrd();
          }
          if (!baseOrd) {
            var ords = getOrdsFromDrag(window.location.href, "text");
            baseOrd = ords[0];
          }

          baseOrd = exportUtil.getOrdWithoutView(baseOrd);

          actionInfo.add({
            slot: 'baseOrd',
            value: baseOrd
          });

          var tab = new Tab({
                item: new SimplePropertySheet(),
                value: actionInfo, //default formalize exportSettings
                title: webChartUtil.lex.get("webChart.ExportWizardCommand.displayName"),
                modified: function(modifiedActionInfo) {//
                  that.updateOptions(modifiedActionInfo, tab);
                }
              });

          tab.loaded = function() {//
            that.updateOptions(actionInfo, tab);
          };


          setDisplayNames(actionInfo);

          return tab;
        });
    };

    ExportWizardCommand.prototype.updateOptions = function(actionInfo, tab) {
      var sheet = tab.widget,
          print = getChoice(actionInfo, "destination") === "print",
          toStation = getChoice(actionInfo, "destination") === "station",
          toCsv = getChoice(actionInfo, "fileType") === "csv",
          relativeOrd = getChoice(actionInfo, "ordType") === "relative";
      if(!sheet.jq()){
        return;
      }

      sheet.$getRowForKey('fileType').jq().toggle(!print);

      sheet.$getRowForKey('viewOnExport').jq().toggle(toStation);

      sheet.$getRowForKey('statusColumn').jq().toggle(toCsv);

      sheet.$getRowForKey('ordType').jq().toggle(!toCsv && !print);
      sheet.$getRowForKey('baseOrd').jq().toggle(relativeOrd && !toCsv && !print);

      return actionInfo;
    };

    ExportWizardCommand.prototype.handleRead = function(tab) {
      var that = this,
          widget = that.$widget;
      return tab.read()
        .then(function(actionInfo) {
          var fileName = actionInfo.get("fileName"),
              fileType = getChoice(actionInfo, "fileType"),
              destination = getChoice(actionInfo, "destination"),
              fullFileName = fileName + "." + fileType,
              checkFilePath,
              saveFilePath,
              lex = webChartUtil.bajaLex,
              webChartLex = webChartUtil.lex,
              lexText;

          if((destination === "download" && !webChartUtil.isValidName(fileName)) ||
            (destination === "station" && !webChartUtil.isValidPath(fileName))){
            lexText = lex.get("IllegalNameException.name", fileName);
            widget.jq().trigger(chartEvents.DISPLAY_ERROR, lexText);
            return Promise.reject(lexText);
          }

          if(destination === "print"){
            setTimeout(function() {
              window.print();
            }, 100);
            return;
          }

          if(destination === "station"){
            checkFilePath = "/file/charts/" + fullFileName;
            saveFilePath = "/webChart/file/save/charts/" + fullFileName;
            //TODO: file path chooser when file space is supported in nav tree
            return webChartUtil.checkPermissions("file:^charts/" + fullFileName, baja.Permissions.ADMIN_WRITE)
              .then(function() {
                return exportUtil.confirmServerFileWrite(checkFilePath);
              })
              .then(function(confirmed) {
                if(confirmed){

                  return Promise.resolve(widget.makeExport(fileType, actionInfo))
                    .then(function(content) {
                      return webChartUtil.ajax(saveFilePath, {
                        type: "POST",
                        contentType: that.getMimeType(fileType),
                        data: content
                      });
                    })
                    .then(function() {
                      if(getChoice(actionInfo, "viewOnExport")){

                        var hyperlinkLocation = "file:^charts/" + fullFileName,
                            isWb = window.niagara &&
                              window.niagara.env &&
                              window.niagara.env.type === "wb";

                        if(!isWb){
                          hyperlinkLocation = "/ord?" + webChartUtil.safeOrd(hyperlinkLocation);
                        }
                        //http://localhost/ord?file:^charts/BigRamp.chart%7Cview:webChart:ChartWidget
                        if(fileType === "chart"){
                          if(window.niagara && window.niagara.env && window.niagara.env.type && window.niagara.env.type === 'mobile'){
                            hyperlinkLocation += "|view:mobile:MobileHistoryAppChartView";
                          }else {
                            hyperlinkLocation += "|view:webChart:ChartWidget";
                          }
                        }else if(fileType === "csv" && !isWb){
                          hyperlinkLocation = checkFilePath; //use file servlet in browser
                        }

                        if(window.niagara && window.niagara.env && isWb){
                          window.niagara.env.hyperlink(hyperlinkLocation);
                        }else{
                          //NCCB-9359: bajaux: niagara.env.hyperlink does not respect view ords for files
                          window.location.assign(hyperlinkLocation);
                        }
                      }
                    });
                }else{
                  return Promise.reject("");
                }
              })
              .catch(function(err) {
                if(err){
                  lexText = webChartLex.get("webChart.exportFailure", fullFileName, err);
                  widget.jq().trigger(chartEvents.DISPLAY_ERROR, lexText);
                  baja.error(err);
                }
              });
          }

          if(destination === "download"){
            // Download the chart.

            Promise.resolve(widget.makeExport(fileType, actionInfo))
              .then(function(content) {
                that.download(content, fileType, fullFileName);
              }).catch(baja.error);
          }
        })
        .catch(function(err) {
          baja.error(err);
          throw err;
        });
    };

    ExportWizardCommand.prototype.download = function(content, fileType, fileName) {
      var that = this,
          blob = webChartUtil.createBlob([content], {type: that.getMimeType(fileType)}),
          a = $("<a/>")
            .attr("href", URL.createObjectURL(blob))
            .attr('download', fileName);

      //IE11 requires this; Edge is uses this too because it ends up with a uuid as the fileName for the a.download
      if(window.navigator && (typeof window.navigator.msSaveBlob === 'function')) {
        window.navigator.msSaveBlob(blob, fileName);
      }
      else
      {
        webChartUtil.click(a);
      }
    };

    ExportWizardCommand.prototype.getMimeType = function(fileType) {
      return this.$defaultMimeTypes[fileType];
    };


    /**
     * Override to set lexicon parameters
     * @param obj
     * @returns {{}}
     */
    ExportWizardCommand.prototype.lexParams = function(obj) {

      return {
        module: "webChart",
        lex: "webChart.ExportWizardCommand"
      };
    };

    ExportWizardCommand.prototype.options = function() {
      return baja.Facets.make();
    };

    return ExportWizardCommand;
  }
)
;
