/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/*jshint browser: true, devel: true */

/**
 * Connection Management for a browser.
 *
 * @module env/ConnectionManager
 * @private
 */
define(["bajaScript/comm",
        "bajaScript/env/HttpOnlyConnection",
        "bajaScript/env/WorkbenchConnection", 
        "bajaScript/env/WebSocketConnection", 
        "bajaScript/env/ReusedWebSocketConnection",
        "bajaPromises"], function (
        baja,
        HttpOnlyConnection,
        WorkbenchConnection,
        WebSocketConnection,
        ReusedWebSocketConnection,
        bajaPromises) {

  "use strict";

  /**
   * Manages full duplex connections for the browser environment.
   *
   * @private
   * @class
   * @param {Function} processNext A function called to process the next frame.
   */
  var ConnectionManager = function (processNext) {
    var that = this,
        sentFrames = that.$sentFrames = [];

    that.$connection = null;
    that.$handlers = {
      
     /**
      * A handler called when a connection is closed.
      *
      * @private
      * 
      * @param [ev] An optional event object.
      */
      close: function (ev) {
        var i;

        if (!baja.isStopping() && !baja.isStopped()) {
          // If there are any outstanding messages then fail them.
          for (i = 0; i < sentFrames.length; ++i) {
            sentFrames[i].cb.fail(ev || "Unknown");
          }
          sentFrames.splice(0, sentFrames.length);
        }

        // Remove BajaScript from event mode.
        baja.comm.stopEventMode();

        // Call to process the next frame.
        processNext();
      },

      /**
       * A handler called when a new message is received from a connection.
       *
       * @private
       * 
       * @param ev An event object.
       */
      message: function (ev) {
        // Parse the response message into a BOX frame
        var boxFrame = JSON.parse(ev.data),
            i,
            matchedFrame,
            frameData;
      
        // Bail if not a BOX frame
        if (!boxFrame || boxFrame.p !== "box") {
          baja.error("Invalid BOX frame: " + ev.data);
          return;
        }

        if (typeof boxFrame.n === "number") {
          // Find a match for the incoming frame.
          for (i = 0; i < sentFrames.length; ++i) {

            frameData = sentFrames[i];
            if (frameData.frame.getSequenceNum() === boxFrame.n) {
              try {                 
                // Make OK callback
                frameData.cb.ok(boxFrame);              
              }
              finally {
                matchedFrame = true;
                sentFrames.splice(i, 1);

                if (frameData.processed) {
                  frameData.processed();
                }

                break;
              }
            }

          }
        }
    
        if (!matchedFrame) {
          // Process unsolictied BOX frames
          for (i = 0; i < boxFrame.m.length; ++i) {
            // Handle unsolicited events
            if (boxFrame.m[i].t === "u") {
              baja.comm.handleEvents(boxFrame.m[i].b.d);
            }
          }
        }
      },

      /**
       * A handler called when a connection is opened.
       *
       * @private
       */
      open: function () {
        // Put BajaScript into event mode.
        baja.comm.startEventMode();
      },

      /**
       * A handler called when a connection goes into error.
       *
       * @private
       * 
       * @param [ev] An optional error object.
       */
      error: function(ev) {
        baja.error(ev);
      }
    };
  };
  
  ConnectionManager.prototype.$makeConnection = function (params) {
    var handlers = this.$handlers,
      hasWebSockets = typeof WebSocket === 'function';
    
    if (window.niagara_wb_util_sendBox) {
      return new WorkbenchConnection(handlers);
    } else if (hasWebSockets && params) {
      return new ReusedWebSocketConnection(handlers, params);
    } else if (hasWebSockets) {
      return new WebSocketConnection(handlers);
    } else {
      return new HttpOnlyConnection();
    }
  };

  /**
   * Starts the connection or short circuits if a connection is already available.
   *
   * @private
   * 
   * @param [params] Optional parameters used to create the underlying connection.
   * @returns {Promise} promise to be resolved when the connection is
   * authenticated and opened.
   */
  ConnectionManager.prototype.start = function (params) {
    var connection = this.$connection;

    if (connection) {
      return bajaPromises.resolve();
    }
  
    // TODO: We should list supported protocols in the WebSocket constructor?
    // https://developer.mozilla.org/en-US/docs/WebSockets/Writing_WebSocket_client_applications
    
    connection = this.$connection = this.$makeConnection(params);
    
    return bajaPromises.resolve(connection.authenticate())
      .then(function () {
        return connection.start();
      });
  };

  /**
   * Send the data to the connection.
   *
   * @private
   * 
   * @param frameData The frame data to send.
   * @returns {Boolean} returns true if successfully sent or false if it can't be sent.
   * If false is returned then the message can't be sent. It may be resent by another medium
   * (i.e. synchronously via HTTP or some other medium).
   */
  ConnectionManager.prototype.send = function (frameData) {
    var that = this;

    // If WebSockets aren't support or the frame
    // needs to be sent synchronously or the web socket isn't open yet
    // then send via HTTP.
    if (frameData.frame.$sync || 
      (!that.$connection || !that.$connection.isOpen())) {
      return false;
    }

    // Remember the BOX frame we've sent so we can reference it on the response.
    that.$sentFrames.push(frameData);
                                 
    // Send the JSON message through the WebSocket
    that.$connection.send(frameData);
           
    // Return true so we can send it
    return true;
  };

  /**
   * Close any underlying connections.
   *
   * @private
   */
  ConnectionManager.prototype.close = function () {
    if (this.$connection) {
      this.$connection.close();
    }
  };

  /**
   * Returns the underlying internal connection object or null if it doesn't exist.
   *
   * @private
   * 
   * @returns The internal connection or null if there isn't one.
   */
  ConnectionManager.prototype.getInternalConnection = function () {
    return this.$connection && this.$connection.getInternalConnection();
  };

  return ConnectionManager;
});