/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/*jshint browser: true, devel: true */

/**
 * Reused WebSocket Connection.
 *
 * @module env/ReusedWebSocketConnection
 * @private
 */
define(["bajaScript/comm",
        "bajaScript/baja/comm/Callback",
        "bajaScript/env/Connection"], function (
        baja,
        Callback,
        Connection) {

  "use strict";

  /**
   * A connection that reuses an existing WebSocket for a connection.
   *
   * @class
   * @private
   * @extends Connection
   * @param {Object} handlers Holds a collection of functions used as asynchronous callbacks.
   * @param {module:env/Connection} internalConnection the existing parent
   * connection to reuse
   */
  var ReusedWebSocketConnection = function (handlers, internalConnection) {
    baja.callSuper(ReusedWebSocketConnection, this, arguments);

    if (!internalConnection) {
      throw new Error("No connection to reuse!");
    }

    this.$internalConnection = internalConnection;
  };

  baja.subclass(ReusedWebSocketConnection, Connection);

  /**
   * Authenticates by creating a new server session.
   * @returns {Promise}
   */
  ReusedWebSocketConnection.prototype.doAuthenticate = function () {
    var cb = new baja.comm.Callback();
    baja.comm.makeServerSession(cb);
    return cb.promise();
  };

  /**
   * Piggyback onto all the event handlers from the parent connection.
   */
  ReusedWebSocketConnection.prototype.start = function () {
    var that = this,
        connection = that.$internalConnection,
        handlers = that.$handlers,
        name;
     
    for (name in handlers) {
      if (handlers.hasOwnProperty(name)) {
        connection.addEventListener(name, handlers[name], false);  
      }
    }

    // Check current state to see if the Socket is already open.
    if (connection.readyState === Connection.WS_OPEN) {
      handlers.open();
    }
  };

  ReusedWebSocketConnection.prototype.close = function () {
    var that = this,
        connection = that.$internalConnection,
        handlers = that.$handlers,
        name;

    for (name in handlers) {
      if (handlers.hasOwnProperty(name)) {
        connection.removeEventListener(name, handlers[name], false);  
      }
    }

    that.$handlers.close();
  };

  return ReusedWebSocketConnection;
});  