/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * @module nmodule/webChart/rc/export/exportUtils
 * @private
 */
define([
  'd3',
  'jquery',
  'Promise',
  'baja!',
  'lex!',
  'bajaux/Widget',
  'dialogs',
  'moment',
  'nmodule/webChart/rc/menu/MenuWidget',
  'bajaux/commands/Command',
  'nmodule/webChart/rc/webChartUtil',
  'nmodule/webChart/rc/dragUtil',
  'nmodule/webChart/rc/choiceUtil',
  'nmodule/webChart/rc/localeUtil'
], function(d3, $, Promise, baja, lex, Widget, dialogs, moment, MenuWidget, Command, webChartUtil, dragUtil, choiceUtil, localeUtil){
  "use strict";

  /**
   * A set of utility functions for exporting
   * @exports nmodule/webChart/rc/export/exportUtil
   */
  var exportUtil = {},
    getOrdsFromDrag = dragUtil.getOrdsFromDrag;

  exportUtil.confirmServerFileWrite = function(filePath){
    return new Promise(function(resolve, reject){
      //TODO: always wrap jquery promise in a Real Promis
      Promise.resolve($.ajax(filePath, {type: "GET"}))
        .then(function fileExists(){
            return lex.module("webChart")
              .then(function(lex){
                return dialogs.showYesNo({
                  title: lex.get("webChart.ExportCommand.overwrite"),
                  content: lex.get("webChart.ExportCommand.askOverwrite")
                }).promise();
              });
          },
          function doesNotExist(){
            // If the file doesn't exist then we're good and
            // we can write the file to the Server.
            resolve(true);
          })
        .spread(function clickedDlg(dlg, id){
          // If the user has clicked 'yes' then we can create the file
          // on the Server.
          resolve(id === "yes");
        });
    });
  };
  /**
   * Export a Widget to csv
   * @param widget Widget to export
   * @param optionInfo Any Options for the export
   * @returns {Promise}
   */
  exportUtil.exportToCsv = function(widget, optionInfo){
    var model          = widget.model(),
        exportObject   = [],
        rowHeader      = [],
        i,
        series,
        dataRow,
        minTimestamp   = Number.MAX_VALUE,
        pointsIndex    = [],
        seriesList     = [],
        allPoints      = [],
        points,
        fullSeriesList = model.seriesList(),
        promises       = [],
        statusColumn   = choiceUtil.getChoice(optionInfo, "statusColumn"),
        columnHeader,
        displayUnitSymbol,
        displayName;

    exportObject[0] = rowHeader;

    // Prepend a byte order mark to indicate we're using UTF-8
    // Certain applications won't recognize it as UTF-8 otherwise,
    // and may contain garbage characters.

    rowHeader.push("\uFEFF" + webChartUtil.lex.get("timestamp"));

    //setup header like HistoryChart->export data
    for(i = 0; i < fullSeriesList.length; i++){
      series = fullSeriesList[i];
      if(!series.isEnabled()){
        continue; //disabled series not included
      }
      seriesList.push(series);
      displayName = series.shortDisplayName();
      columnHeader = displayName;
      displayUnitSymbol = series.valueScale().displayUnitSymbol();
      if(displayUnitSymbol){
        columnHeader += "(" + displayUnitSymbol + ")"; //TODO: excel fix
      }
      rowHeader.push(columnHeader);

      if(statusColumn){
        rowHeader.push(displayName + " (" + webChartUtil.lex.get("status") + ")");
      }
      points = series.samplingPoints();
      allPoints.push(points);
      pointsIndex.push(0);
    }

    //gets points ready for comparison
    minTimestamp = exportUtil.nextTimestamp(minTimestamp, allPoints, pointsIndex, seriesList);

    while(true){
      dataRow = exportUtil.dataRow(minTimestamp, allPoints, pointsIndex, seriesList, promises, statusColumn);
      if(dataRow){
        exportObject.push(dataRow);
      }else{
        break;
      }
      minTimestamp = exportUtil.nextTimestamp(minTimestamp, allPoints, pointsIndex);
    }

    return Promise.all(promises)
      .then(function(){
        return d3.csv.format(exportObject);
      });
  };

  exportUtil.resolveData = function(dataRow, index, y, series, status, statusColumn){
    return series.resolveYDisplay(y)
      .then(function(display){
        dataRow[index] = display;
        if(statusColumn){
          return webChartUtil.statusToString(status)
            .then(function(statusString){
              dataRow[index + 1] = statusString;
            });
        }
      });
  };

  exportUtil.dataRow = function(timestamp, allPoints, pointsIndex, seriesList, promises, statusColumn){
    var dataRow      = [],
        i,
        pointIndex,
        points,
        point,
        useful       = false,
        dataRowIndex = 0;


    for(i = 0; i < allPoints.length; i++){
      points = allPoints[i];
      pointIndex = pointsIndex[i];
      point = points[pointIndex];
      if(point && !point.skip && timestamp === point.x.getTime()){

        if(statusColumn){
          dataRowIndex = i * 2 + 1;
        }else{
          dataRowIndex = i + 1;
        }
        promises.push(exportUtil.resolveData(dataRow, dataRowIndex, point.y, seriesList[i], point.status, statusColumn));

        useful = true;
        pointsIndex[i]++; //pointIndex for series is used, so increment by 1
      }
      else{
        //leave blank for now
        //dataRow[i+1] = "nan";
      }
    }

    if(useful){
      dataRow[0] = localeUtil.format(localeUtil.getD3DateTimeFormat(), new Date(timestamp));
      return dataRow;
    }else{
      return null;
    }
  };

  exportUtil.nextTimestamp = function(timestamp, allPoints, pointsIndex){
    var i,
        pointIndex,
        points,
        point,
        nextTimestamp = Number.MAX_VALUE;


    for(i = 0; i < allPoints.length; i++){
      points = allPoints[i];
      pointIndex = pointsIndex[i];
      point = points[pointIndex];
      while(point && point.skip){
        pointsIndex[i]++;
        pointIndex = pointsIndex[i];
        point = points[pointIndex];
      }

      if(point && !point.skip){
        nextTimestamp = Math.min(nextTimestamp, point.x.getTime());
      }
    }
    return nextTimestamp;
  };

  /**
   * Relativize a path
   *
   * @param {Array<string>} from base path
   * @param {Array<string>} to path being relativized
   * @returns {string}
   */
  exportUtil.relative = function(from, to) {
    var names = [], i, a = 0, b;

    // find matching section
    while ((a < to.length) && (a < from.length) && (from[a].equals(to[a]))) {
      a++;
    }

    // back up
    b = from.length - a;
    for (i = 0; i < b; i++)
    {
      if (names.length > 0) {
        names.push("/");
      }
      names.push("..");
    }

    // go forwards
    for (i = a; i < to.length; i++)
    {
      if (names.length > 0) {
        names.push("/");
      }
      names.push(to[i]);
    }

    return names.join('');
  };

  /**
   * Convert an ord string to an ord string with the view query removed.
   *
   * @param ordString
   * @returns {string}
   */
  exportUtil.getOrdWithoutView = function(ordString) {
    return ordString.replace(/\|view:.*/, '');
  };

  /**
   * Get the index of the first relevant item in the OrdQueryList.
   *
   * Skips station, memory, and bog.
   *
   * @param {baja.OrdQueryList} q
   * @returns {Number}
   */
  exportUtil.getPathIndex = function(q) {
    var stationFound,
        n;

    // Skip past station: scheme if there is one.
    for(n = 0; n < q.size(); n++) {
      var schemeName = q.get(n).getSchemeName();
      if(schemeName === "memory") {
        stationFound = true;
        n = n + 2;
        break;
      }
      if(schemeName === "station" || schemeName === "bog") {
        stationFound = true;
        n = n + 1;
        break;
      }
    }
    if(!stationFound) {
      n = 0;
    }
    return n;
  };

  /**
   * Relativize an Ord
   *
   * This logic is duplicate to pxEditor:RelativizeOrds
   *
   * @param {baja.Ord} baseOrd
   * @param {baja.Ord} oldOrd
   * @returns {baja.Ord}
   */
  exportUtil.relativizeOrd = function(baseOrd, oldOrd) {
    var q = oldOrd.normalize().parse();
    var n = exportUtil.getPathIndex(q);

    // If the first query is not a slot path, just return the original ord.
    var path = q.get(n);

    if(path.getSchemeName() !== "slot") {
      return oldOrd;
    }

    var names = path.getNames();

    var baseOrdQueryList = baseOrd.normalize().parse();
    var basePath = baseOrdQueryList.get(exportUtil.getPathIndex(baseOrdQueryList));

    if(basePath.getSchemeName() !== "slot") {
      return oldOrd;
    }
    var baseNames = basePath.getNames();

    var rel = exportUtil.relative(baseNames, names);
    var queries = [];
    queries.push(new baja.SlotPath(rel));
    for(var i = n + 1; i < q.length; i++) {
      queries.push(q[i]);
    }

    return baja.Ord.make(queries.join(''));
  };

  /**
   * Get the current base ORD; e.g. "local:|foxs:|station:|slot:/Folder".
   *
   * @returns {string}
   */
  exportUtil.getBaseOrd = function () {
    // Get base folder / ord
    var baseOrd;
    if (typeof window.niagara !== 'undefined' &&
      window.niagara.env &&
      typeof window.niagara.env.getBaseOrd === 'function') {
      baseOrd = window.niagara.env.getBaseOrd();
    }
    if (!baseOrd) {
      var ords = getOrdsFromDrag(window.location.href, "text");
      baseOrd = ords[0];
    }

    baseOrd = exportUtil.getOrdWithoutView(baseOrd);
    return baseOrd;
  };

  return exportUtil;
});
