/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Jeremy Narron
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
 */
define(['lex!webEditors',
        'bajaux/Widget',
        'bajaux/commands/Command',
        'jquery',
        'Promise',
        'nmodule/js/rc/switchboard/switchboard',
        'nmodule/webEditors/rc/fe/CompositeEditor',
        'nmodule/webEditors/rc/fe/feDialogs',
        'nmodule/webEditors/rc/fe/baja/DisplayOnlyEditor',
        'nmodule/webEditors/rc/fe/array/ArrayCompositeBuilder'], function (
         lexs,
         Widget,
         Command,
         $,
         Promise,
         switchboard,
         CompositeEditor,
         feDialogs,
         DisplayOnlyEditor,
         ArrayCompositeBuilder) {

  'use strict';

  var webEditorsLex = lexs[0];

  ////////////////////////////////////////////////////////////////
  // Support functions
  ////////////////////////////////////////////////////////////////

  function swap(arr, i, j) { var tmp = arr[i]; arr[i] = arr[j]; arr[j] = tmp; }

  ////////////////////////////////////////////////////////////////
  // ArrayCompositeEditor
  ////////////////////////////////////////////////////////////////

  /**
   * An abstract editor intended to manage the instantiation, initialization,
   * and loading of multiple sub-editors contained in an array.
   *
   * This class also contains default implementations of four commands
   * that can be used as buttons with this editor: add, remove, move up,
   * and move down.
   *
   * It accepts an object literal Property called `numbers` which contains two
   * sub properties:
   * - `show`: (Boolean) if true display the number for each editor starting
   * from 1.
   * - `countFromZero`: (Boolean) if true start the editor numbers from 0
   * instead of 1.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/CompositeEditor
   * @alias module:nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
   */
  var ArrayCompositeEditor = function ArrayCompositeEditor(params) {
    CompositeEditor.apply(this, arguments);

    if (this.getFormFactor() === Widget.formfactor.mini) {
      DisplayOnlyEditor.$mixin(this);
    }

    switchboard(this, { doLoad: { allow: 'oneAtATime', onRepeat: 'queue' } });
  };
  ArrayCompositeEditor.prototype = Object.create(CompositeEditor.prototype);
  ArrayCompositeEditor.prototype.constructor = ArrayCompositeEditor;

  /**
   * Configures the array composite builder for the loaded value.
   * This will create one editor for each value configured.
   *
   * @returns {module:nmodule/webEditors/rc/fe/array/ArrayCompositeBuilder}
   */
  ArrayCompositeEditor.prototype.makeBuilder = function () {
    var that = this,
        builder = new ArrayCompositeBuilder(),
        destroyFor = builder.destroyFor;

    builder.getDomFor = function (i) {
      var numbers = that.properties().getValue('numbers') ||
        { show: true, countFromZero: false };

      var parent = $('<div class="editor-row"/>');
      var editor = $('<div/>');
      parent.appendTo(that.$getEditorsElement());

      if (numbers && numbers.show) {
        var index = numbers && numbers.countFromZero ? Number(i) : Number(i) + 1;
        var number = $('<div class="number">' + index + '.</div>');
        number.appendTo(parent);
      }

      return editor.appendTo(parent);
    };

    builder.destroyFor = function (key) {
      return destroyFor.apply(builder, arguments)
        .then(function () {
          that.$getEditorsElement().children().eq(key).remove();
        });
    };

    return builder;
  };

  /**
   * Arm event handlers for managing the currently-focused editor.
   *
   * @param {JQuery} dom
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.doInitialize = function (dom) {
    var that = this;

    dom.addClass('ArrayCompositeEditor');
    dom.append('<div class="editors"/>');

    dom.on('focus', '.editors > .editor-row', function () {
      that.$setFocusedEditorIndex($(this).index());
    });

    return CompositeEditor.prototype.doInitialize.apply(this, arguments);
  };

  /**
   * Read a new value for the currently entered value.
   *
   * @returns {Promise.<Array>}
   */
  ArrayCompositeEditor.prototype.doRead = function () {
    return this.getBuilder().readAll();
  };

  /**
   * Destroy all child editors.
   *
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.doDestroy = function () {
    this.jq().removeClass('ArrayCompositeEditor');
    return this.getChildEditors().destroyAll();
  };

  /**
   * Set all child editors readonly/writable.
   *
   * @param {Boolean} readonly
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.doReadonly = function (readonly) {
    return this.getChildEditors().setAllReadonly(readonly);
  };

  /**
   * Set all child editors enabled/disabled.
   *
   * @param {Boolean} enabled
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.doEnabled = function (enabled) {
    return this.getChildEditors().setAllEnabled(enabled);
  };

  /**
   * @private
   * @returns {jQuery}
   */
  ArrayCompositeEditor.prototype.$getEditorsElement = function () {
    return this.jq().children('.editors');
  };

  /**
   * Gets the currently focused editor index.
   *
   * @returns {Number}
   */
  ArrayCompositeEditor.prototype.getFocusedEditorIndex = function () {
    return this.$focusedEditorIndex;
  };

  /**
   * Sets the currently focused editor index.
   *
   * @private
   * @param {Number} i
   * @param {Boolean} requestFocus
   */
  ArrayCompositeEditor.prototype.$setFocusedEditorIndex = function (i, requestFocus) {
    var focusedEd = this.getBuilder().getEditorFor(String(i));
    this.$focusedEditorIndex = i;
    this.$getEditorsElement().children()
      .removeClass('ux-bg')
      .eq(i).addClass('ux-bg');
    if (i !== undefined) {
      if (requestFocus) {
        focusedEd.requestFocus();
      }
    }
  };

  /**
   * When in mini/display-only mode, just display the number of items.
   *
   * @returns {String}
   */
  ArrayCompositeEditor.prototype.valueToString = function (array) {
    return webEditorsLex.get('ArrayCompositeEditor.items', array.length);
  };

  ////////////////////////////////////////////////////////////////
  // Command helper functions
  ////////////////////////////////////////////////////////////////

  /**
   * Add an item to the backing array.
   *
   * @param {*} item
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.addItem = function (item) {
    var array = this.value();
    array.push(item);
    return this.doLoad(array);
  };

  /**
   * Remove an item from the backing array.
   *
   * @param {Number} index
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.removeItem = function (index) {
    var array = this.value();

    if(index < 0 || index >= array.length) {
      return Promise.reject(new Error('Index out of array bounds'));
    }

    array.splice(index, 1);

    var focusedIndex = this.getFocusedEditorIndex();

    if (index === focusedIndex) {
      if(index === this.$getEditorsElement().children().length - 1) {
        this.$setFocusedEditorIndex(undefined);
      }
    } else if (focusedIndex > index) {
      this.$setFocusedEditorIndex(focusedIndex - 1, true);
    }

    return this.doLoad(array);
  };

  /**
   * Move an item one slot up in the backing array.
   *
   * @param {Number} index
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.moveUpItem = function (index) {
    var array = this.value();

    if(index < 0 || index >= array.length) {
      return Promise.reject(new Error('Index out of array bounds'));
    }

    if (index === 0) { return Promise.resolve(); }
    swap(array, index, index - 1);

    if (index === this.getFocusedEditorIndex()) {
      this.$setFocusedEditorIndex(index - 1, true);
    }

    return this.doLoad(array);
  };

  /**
   * Move an item one slot down in the backing array.
   *
   * @param {Number} index
   * @returns {Promise}
   */
  ArrayCompositeEditor.prototype.moveDownItem = function (index) {
    var array = this.value();

    if(index < 0 || index >= array.length) {
      return Promise.reject(new Error('Index out of array bounds'));
    }

    if (index === array.length - 1) { return Promise.resolve(); }
    swap(array, index, index + 1);

    if (index === this.getFocusedEditorIndex()) {
      this.$setFocusedEditorIndex(index + 1, true);
    }

    return this.doLoad(array);
  };

  ////////////////////////////////////////////////////////////////
  // Commands
  ////////////////////////////////////////////////////////////////

  /**
   * Command to add a new String.
   *
   * Use a different Command if you need to add something other than a String
   * to your ArrayCompositeEditor.
   *
   * @private
   * @class
   * @extends module:bajaux/commands/Command
   * @memberOf module:nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
   * @param {Function} getEd a function that returns either an
   * `ArrayCompositeEditor` or a `Promise` that resolves with an
   * `ArrayCompositeEditor`.
   */
  var AddCommand = function AddCommand(getEd) {
    Command.call(this, {
      module: 'webEditors',
      lex: 'commands.addItem',
      func: function () {
        return Promise.join(
          feDialogs.showFor({ value: '' }),
          getEd()
        )
        .spread(function (item, ed) {
          if (item === null) { return; }
          return ed.addItem(item);
        });
      }
    });
  };
  AddCommand.prototype = Object.create(Command.prototype);
  AddCommand.prototype.constructor = AddCommand;
  ArrayCompositeEditor.AddCommand = AddCommand;

  /**
   * Command to remove the selected item.
   *
   * @private
   * @class
   * @extends module:bajaux/commands/Command
   * @memberOf module:nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
   * @param {Function} getEd a function that returns either an
   * `ArrayCompositeEditor` or a `Promise` that resolves with an
   * `ArrayCompositeEditor`.
   */
  var RemoveCommand = function RemoveCommand(getEd) {
    Command.call(this, {
      module: 'webEditors',
      lex: 'commands.removeItem',
      func: function () {
        return Promise.resolve(getEd())
          .then(function (ed) {
            var focusedIndex = ed.getFocusedEditorIndex();
            if (focusedIndex !== undefined) {
              return ed.removeItem(focusedIndex);
            }
        });
      }
    });
  };
  RemoveCommand.prototype = Object.create(Command.prototype);
  RemoveCommand.prototype.constructor = RemoveCommand;
  ArrayCompositeEditor.RemoveCommand = RemoveCommand;

  /**
   * Command to move the selected item up in the list.
   *
   * @private
   * @class
   * @extends module:bajaux/commands/Command
   * @memberOf module:nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
   * @param {Function} getEd a function that returns either an
   * `ArrayCompositeEditor` or a `Promise` that resolves with an
   * `ArrayCompositeEditor`.
   */
  var MoveUpCommand = function MoveUpCommand(getEd) {
    Command.call(this, {
      module: 'webEditors',
      lex: 'commands.moveUp',
      func: function () {
        return Promise.resolve(getEd())
          .then(function (ed) {
            var focusedIndex = ed.getFocusedEditorIndex();
            if (focusedIndex !== undefined) {
              return ed.moveUpItem(focusedIndex);
            }
        });
      }
    });
  };
  MoveUpCommand.prototype = Object.create(Command.prototype);
  MoveUpCommand.prototype.constructor = MoveUpCommand;
  ArrayCompositeEditor.MoveUpCommand = MoveUpCommand;

  /**
   * Command to move the selected item down in the list.
   *
   * @private
   * @class
   * @extends module:bajaux/commands/Command
   * @memberOf module:nmodule/webEditors/rc/fe/array/ArrayCompositeEditor
   * @param {Function} getEd a function that returns either an
   * `ArrayCompositeEditor` or a `Promise` that resolves with an
   * `ArrayCompositeEditor`.
   */
  var MoveDownCommand = function MoveDownCommand(getEd) {
    Command.call(this, {
      module: 'webEditors',
      lex: 'commands.moveDown',
      func: function () {
        return Promise.resolve(getEd())
          .then(function (ed) {
            var focusedIndex = ed.getFocusedEditorIndex();
            if (focusedIndex !== undefined) {
              return ed.moveDownItem(focusedIndex);
            }
        });
      }
    });
  };
  MoveDownCommand.prototype = Object.create(Command.prototype);
  MoveDownCommand.prototype.constructor = MoveDownCommand;
  ArrayCompositeEditor.MoveDownCommand = MoveDownCommand;

  return ArrayCompositeEditor;
});
