/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/BooleanEditor
 */
define(['baja!',
        'lex!baja',
        'log!nmodule.webEditors.rc.fe.baja.BooleanEditor',
        'jquery',
        'Promise',
        'nmodule/webEditors/rc/fe/baja/BaseEditor'], function (
         baja,
         lexs,
         log,
         $,
         Promise,
         BaseEditor) {
  'use strict';

  var bajaLex = lexs[0],
      logError = log.severe.bind(log);

  function doFormat(pattern) {
    return baja.Format.format({ pattern: pattern });
  }

  /**
   * A field editor for working with booleans.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/BooleanEditor
   */
  var BooleanEditor = function BooleanEditor(params) {
    BaseEditor.call(this, $.extend({
      keyName: 'BooleanEditor'
    }, params));
  };
  BooleanEditor.prototype = Object.create(BaseEditor.prototype);
  BooleanEditor.prototype.constructor = BooleanEditor;

  /**
   * Compute the true/false display text to use from the `trueText` /
   * `falseText` facets.
   *
   * @private
   * @returns {Promise} promise to be resolved with an array containing two
   * strings for trueText/falseText
   */
  BooleanEditor.prototype.$getTrueFalseText = function () {
    var props = this.properties(),
        trueText = props.getValue('trueText') || bajaLex.get('true'),
        falseText = props.getValue('falseText') || bajaLex.get('false');
    return Promise.join(doFormat(trueText), doFormat(falseText));
  };

  /**
   * Return the checkbox element.
   *
   * @private
   * @returns {jQuery}
   */
  BooleanEditor.prototype.$getCheckbox = function () {
    return this.jq().children('input');
  };

  /**
   * Update the current value of the display label with the `trueText` or
   * `falseText` facet, falling back to `true` and `false` from the `baja`
   * lexicon if either of those facets is not configured.
   * @private
   * @param {Boolean} value whether the the editor currently has true or false
   * loaded
   * @returns {Promise}
   */
  BooleanEditor.prototype.$updateLabel = function (value) {
    var that = this;
    return that.$getTrueFalseText()
      .spread(function (trueText, falseText) {
        that.jq().children('label').text(value ? trueText : falseText);
      });
  };

  /**
   * Creates a checkbox input.
   * @param {JQuery} dom
   */
  BooleanEditor.prototype.doInitialize = function (dom) {
    var that = this,
        id = this.generateId(),
        input = $('<input type="checkbox">')
          .prop('id', id),
        label = $('<label></label>')
          .prop('for', id);
    dom.html(input.add(label));
    dom.on('change', 'input:checkbox', function () {
      that.setModified(true);
      that.read()
        .then(function (value) {
          return that.$updateLabel(value);
        })
        .catch(logError);
    });
  };

  /**
   * Checks or unchecks the checkbox to reflect the loaded value.
   * @param {Boolean} val
   */
  BooleanEditor.prototype.doLoad = function (val) {
    this.$getCheckbox().prop('checked', !!val);
    return this.$updateLabel(val);
  };

  /**
   * Returns the checked/unchecked state of the checkbox.
   * @returns {Boolean} the boolean checked state of the checkbox
   */
  BooleanEditor.prototype.doRead = function () {
    return this.$getCheckbox().prop('checked');
  };

  /**
   * Enables or disables the checkbox element.
   *
   * @param {Boolean} enabled
   */
  BooleanEditor.prototype.doEnabled = function (enabled) {
    this.$getCheckbox().prop('disabled', this.isReadonly() || !enabled);
  };

  /**
   * Disables or enables the checkbox element.
   *
   * @param {Boolean} readonly
   */
  BooleanEditor.prototype.doReadonly = function (readonly) {
    this.$getCheckbox().prop('disabled', !this.isEnabled() || readonly);
  };

  return (BooleanEditor);
});
