/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * FileInputEditor provides a reference to a javascript file chosen by the user.
 * Calls to save and read must conform to baja types (which can't be a javascript file) so
 * those calls return the fileName (for informational purposes). The reference to the actual file
 * is obtained via FileInputEditor.getLastUploadedFile() after FileInputEditor.doSave() has been called.
 *
 * @private
 * @module nmodule/webEditors/rc/fe/baja/FileInputEditor
 */
define(['jquery',
        'nmodule/webEditors/rc/fe/baja/BaseEditor'], function (
         $,
         BaseEditor) {
  'use strict';

  var $lastUploadedFile = null;

  /**
   * A field editor for helping upload a file.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/FileInputEditor
   */
  var FileInputEditor = function FileInputEditor(params) {
    BaseEditor.call(this, $.extend({
      keyName: 'FileInputEditor'
    }, params));
  };
  FileInputEditor.prototype = Object.create(BaseEditor.prototype);
  FileInputEditor.prototype.constructor = FileInputEditor;

  /**
   * Return the input element.
   *
   * @private
   * @returns {jQuery}
   */
  FileInputEditor.prototype.$getInput = function () {
    return this.jq().children('input');
  };

  /**
   * Creates a file input
   * @param {JQuery} dom
   */
  FileInputEditor.prototype.doInitialize = function (dom) {
    var that  = this,
        input = $('<input type="file">');
    dom.html(input);
    input.on('change', function (event) {
      that.$updateFile(event);
    });
  };

  /**
   * update the $lastChangedFile from the last change event
   * @param {Object} event
   */
  FileInputEditor.prototype.$updateFile = function (event) {
    var that  = this,
        files = event.target.files;
    that.$lastChangedFile = files[0];
    that.setModified(true);
  };

  /**
   * Returns the file Contents from the last uploaded file that has been Saved.
   * @returns {Object} Javascript file reference
   */
  FileInputEditor.getLastUploadedFile = function () {
    return $lastUploadedFile;
  };

  /**
   * Read the File name from the widget.
   * @returns {String} FileName of Javascript file if its been set
   */
  FileInputEditor.prototype.doRead = function () {
    var that = this;
    if (that.$lastChangedFile) {
      return that.$lastChangedFile.name;
    }
  };

  /**
   * Enables or disables the input element.
   *
   * @param {Boolean} enabled
   */
  FileInputEditor.prototype.doEnabled = function (enabled) {
    this.$getInput().prop('disabled', this.isReadonly() || !enabled);
  };

  /**
   * Disables or enables the input element.
   *
   * @param {Boolean} readonly
   */
  FileInputEditor.prototype.doReadonly = function (readonly) {
    this.$getInput().prop('disabled', !this.isEnabled() || readonly);
  };

  /**
   * Removes the File Content from the Widget.
   */
  FileInputEditor.prototype.doDestroy = function () {
    this.$lastChangedFile = null;
  };

  /**
   * Save the changed file to the static entry available from FileInputEditor.getLastUploadedFile()
   * Saves the fileName as the String returned
   * @return {String}
   */
  FileInputEditor.prototype.doSave = function () {
    var that = this,
        name = that.$lastChangedFile ? that.$lastChangedFile.name : "";

    $lastUploadedFile = that.$lastChangedFile;


    return name;
  };

  return FileInputEditor;
});
