/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/NameListEditor
 */
define(['baja!',
        'bajaux/Widget',
        'bajaux/commands/Command',
        'bajaux/util/CommandButtonGroup',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/CompositeEditor',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/fe/feDialogs',
        'nmodule/webEditors/rc/fe/baja/DisplayOnlyEditor',
        'nmodule/webEditors/rc/fe/config/CompositeBuilder'], function (
         baja,
         Widget,
         Command,
         CommandButtonGroup,
         $,
         Promise,
         _,
         CompositeEditor,
         fe,
         feDialogs,
         DisplayOnlyEditor,
         CompositeBuilder) {

  'use strict';

  var NameList = baja.NameList;

  /**
   * Command to add a new string to the NameList.
   *
   * @inner
   * @class
   * @extends module:bajaux/commands/Command
   * @param {module:nmodule/webEditors/rc/fe/baja/NameListEditor} ed
   */
  var AddCommand = function AddCommand(ed) {
    Command.call(this, {
      module: 'alarm',
      lex: 'commands.addMsg',
      func: function () {
        return feDialogs.showFor({ value: '' })
          .then(function (str) {
            if (!str) { return; }

            return ed.read()
              .then(function (nameList) {
                return ed.doLoad(nameList.union(NameList.make(baja.SlotPath.escape(str))));
              });
          });
      }
    });
  };
  AddCommand.prototype = Object.create(Command.prototype);
  AddCommand.prototype.constructor = AddCommand;

  /**
   * Command to remove the selected name from the name list.
   *
   * @inner
   * @class
   * @extends module:bajaux/commands/Command
   * @param {module:nmodule/webEditors/rc/fe/baja/NameListEditor} ed
   */
  var RemoveCommand = function RemoveCommand(ed) {
    Command.call(this, {
      module: 'alarm',
      lex: 'commands.removeMsg',
      func: function () {
        var focusedEd = ed.$focusedEditor;
        if (focusedEd) {
          return Promise.join(focusedEd.read(), ed.read())
            .spread(function (toRemove, nameList) {
              return ed.doLoad(nameList.difference(NameList.make(toRemove)));
            });
        }
      }
    });
  };
  RemoveCommand.prototype = Object.create(Command.prototype);
  RemoveCommand.prototype.constructor = RemoveCommand;

  /**
   * Editor for handling `baja:NameList` values.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/CompositeEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/NameListEditor
   */
  var NameListEditor = function NameListEditor(params) {
    var that = this;

    CompositeEditor.apply(that, arguments);

    if (that.getFormFactor() === Widget.formfactor.mini) {
      DisplayOnlyEditor.$mixin(that);
    }

    that.getCommandGroup().add(
      new AddCommand(that),
      new RemoveCommand(that)
    );
  };
  NameListEditor.prototype = Object.create(CompositeEditor.prototype);
  NameListEditor.prototype.constructor = NameListEditor;

  /**
   * @private
   * @returns {module:bajaux/commands/Command}
   */
  NameListEditor.prototype.$getAddCommand = function () {
    return this.getCommandGroup().get(0);
  };

  /**
   * @private
   * @returns {module:bajaux/commands/Command}
   */
  NameListEditor.prototype.$getRemoveCommand = function () {
    return this.getCommandGroup().get(1);
  };

  /**
   * @private
   * @returns {module:bajaux/util/CommandButtonGroup}
   */
  NameListEditor.prototype.$getCommandButtonGroup = function () {
    return this.jq().children('.CommandButtonGroup').data('widget');
  };

  /**
   * @private
   * @returns {jQuery}
   */
  NameListEditor.prototype.$getEditorsElement = function () {
    return this.jq().children('.editors');
  };

  /**
   * When in mini/display-only mode, display the unescaped names.
   *
   * @param {baja.NameList} nameList
   * @returns {string}
   */
  NameListEditor.prototype.valueToString = function (nameList) {
    return _.map(nameList.getNames(), baja.SlotPath.unescape).join(', ');
  };

  /**
   * Configures the composite builder for the loaded `NameList`. Keys will
   * correspond to the indexes in the array of all names, and values will be
   * the SlotPath-unescaped names.
   *
   * @returns {module:nmodule/webEditors/rc/fe/config/CompositeBuilder}
   */
  NameListEditor.prototype.makeBuilder = function () {
    var that = this,
        builder = new CompositeBuilder();

    builder.getDomFor = function (i) {
      return $('<div/>').appendTo(that.$getEditorsElement());
    };

    builder.getKeys = function () {
      return _.map(this.getDataSource().getNames(), function (inst, i) {
        return String(i);
      });
    };

    builder.getValueFor = function (i) {
      return baja.SlotPath.unescape(this.getDataSource().getNames()[i]);
    };

    builder.getConfigFor = function () {
      return { formFactor: 'mini' };
    };

    return builder;
  };

  /**
   * Arm event handlers for managing the currently-focused `String` editor, and
   * build a `CommandButtonGroup` to show the editor's commands.
   * @param {JQuery} dom
   */
  NameListEditor.prototype.doInitialize = function (dom) {
    var that = this;

    dom.append('<div class="editors"/>');

    dom.on('focus', '.editors .type-baja-String', function () {
      var $this = $(this);
      that.$getEditorsElement().children().removeClass('ux-bg');
      $this.addClass('ux-bg');
      that.$focusedEditor = $this.data('widget');
    });

    return Promise.join(
      fe.buildFor({
        dom: $('<div class="editor"/>').appendTo(dom),
        type: CommandButtonGroup,
        value: this.getCommandGroup()
      }),
      CompositeEditor.prototype.doInitialize.apply(this, arguments)
    );
  };

  /**
   * Read all entered strings, SlotPath-escape them, and assemble them into a
   * `baja:NameList` instance.
   * @returns {Promise}
   */
  NameListEditor.prototype.doRead = function () {
    return this.getBuilder().readAll()
      .then(function (names) {
        return NameList.make(_.map(names, baja.SlotPath.escape));
      });
  };

  /**
   * Destroy all child editors.
   *
   * @returns {Promise}
   */
  NameListEditor.prototype.doDestroy = function () {
    return this.getChildEditors().destroyAll();
  };

  /**
   * Set all child editors readonly/writable.
   *
   * @param {Boolean} readonly
   * @returns {Promise}
   */
  NameListEditor.prototype.doReadonly = function (readonly) {
    return this.getChildEditors().setAllReadonly(readonly);
  };

  /**
   * Set all child editors enabled/disabled.
   *
   * @param {Boolean} enabled
   * @returns {Promise}
   */
  NameListEditor.prototype.doEnabled = function (enabled) {
    return this.getChildEditors().setAllEnabled(enabled);
  };

  return (NameListEditor);
});
