/*
 * @copyright 2018 Tridium, Inc. All Rights Reserved.
 * @author Andy Sutton
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/PermissionsMapGridEditor
 */
define(['baja!',
        'bajaux/events',
        'jquery',
        'underscore',
        'Promise',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/fe/baja/util/spaceUtils',
        'nmodule/webEditors/rc/fe/baja/util/subscriberUtils',
        'nmodule/webEditors/rc/fe/baja/PermissionsRowEditor',
        'bajaux/mixin/responsiveMixIn',
        'lex!wbutil,bajaui',
        'hbs!nmodule/webEditors/rc/fe/baja/template/PermissionsTable'], function (
         baja,
         events,
         $,
         _,
         Promise,
         BaseEditor,
         fe,
         spaceUtils,
         subscriberUtils,
         PermissionsRowEditor,
         responsiveMixIn,
         lexs,
         tplTable) {

  'use strict';

  var wbutilLex = lexs[0],
    bajauiLex = lexs[1],
    resolveService = spaceUtils.resolveService,
    leaseWhenNeeded = subscriberUtils.leaseWhenNeeded,
    getRoleService = leaseWhenNeeded(_.partial(resolveService, 'baja:RoleService')),
    getCategoryService = leaseWhenNeeded(_.partial(resolveService, 'baja:CategoryService')),
    getUserService = leaseWhenNeeded(_.partial(resolveService, 'baja:UserService')),
    MODIFY_EVENT = events.MODIFY_EVENT,
    MIN_GRID_ROWS = 8,
    PERMISSIONS_GRID_SELECTOR = '.permissionsGridContent';

  /**
   * A field editor for working with a PermissionsMap.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/PermissionsMapGridEditor
   *
   */
  var PermissionsMapGridEditor = function PermissionsMapGridEditor(params) {
    var that = this;

    BaseEditor.call(that, $.extend(true, {
      keyName: 'PermissionsMapGridEditor'
    }, params));

    responsiveMixIn(that, {
      'permissions-grid-mobile': {
        maxWidth: 760
      }
    });

  };

  PermissionsMapGridEditor.prototype = Object.create(BaseEditor.prototype);
  PermissionsMapGridEditor.prototype.constructor = PermissionsMapGridEditor;

  /**
   * The minimum number of permissions rows to show in the permissionsGrid
   * @private
   */
  PermissionsMapGridEditor.MIN_GRID_ROWS = MIN_GRID_ROWS;

  /**
   * Creates a PermissionsMap element.
   *
   * @param {JQuery} dom
   * @returns {Promise}
   */
  PermissionsMapGridEditor.prototype.doInitialize = function (dom) {
    var that = this;

    dom.html(tplTable({
      showRowDescription: true,

      category: wbutilLex.get('category.category'),
      operator: bajauiLex.get('security.operator'),
      admin: bajauiLex.get('security.admin'),
      read: bajauiLex.get('security.r'),
      write: bajauiLex.get('security.w'),
      invoke: bajauiLex.get('security.i')
    }));

    //if a row is modified, set myself modified
    dom.on(MODIFY_EVENT, '.editor', function () {
      that.setModified(true);
      return false;
    });
  };

  /**
   * Loads the given baja.PermissionsMap into the editor.
   *
   * @param {baja.PermissionsMap} permissionsMap
   * @returns {Promise.<baja.PermissionsMap>}
   */
  PermissionsMapGridEditor.prototype.doLoad = function (permissionsMap) {
    var that = this,
      permissionsGrid = that.jq().find(PERMISSIONS_GRID_SELECTOR);

    return that.getChildEditors().destroyAll().then(function () {
      return permissionsGrid.empty();
    })
    .then(function () {
      return Promise.all([
        that.$getMaxPermissions(),
        that.$getCategories(),
        that.$editorEnabled()
      ]).spread(function (maxPermissions, categoryNames, editorEnabled) {
        return Promise.all(categoryNames.map(function (categoryName, index) {
          var perms;

          if (index === 0) { return; } // don't do anything for index 0
                                       // (as we don't populate it in the category array, it shouldn't exist anyway)

          if (index < permissionsMap.size()) {
            perms = permissionsMap.getPermissions(index);
          } else {
            perms = baja.Permissions.none;
          }

          return fe.buildFor({
            dom:  $('<tr class="ux-table-row" />').appendTo(permissionsGrid),
            type: PermissionsRowEditor,
            value: perms,
            rowDescription: categoryNames[index],
            rowId: index,
            maxPermissions: maxPermissions.getPermissions(index),
            formFactor: 'mini',
            enabled: editorEnabled
          });
        }));
      });
    });
  };

  /**
   * Read the individual child editors and assemble them into a `PermissionsMap`
   *
   * @returns {Promise.<baja.PermissionsMap>} promise to be resolved with a `baja.PermissionsMap` value
   */
  PermissionsMapGridEditor.prototype.doRead = function () {
    var that = this,
      permissions = [];

    return that.$editorEnabled().then(function (editorEnabled) {

      // if the edidtor is not enabled return the underlying (unchanged) value
      if (!editorEnabled) { return Promise.resolve(that.value()); }

      return Promise.all(
        that.getChildEditors().map(function(row) {
          return row.read().then(function (perms) {
            permissions[row.getRowId()] = perms;
          });
        })
      ).then(function () {
        return baja.PermissionsMap.make(permissions);
      });
    });
  };

  /**
   * Whether to enable the editor
   * The main use case here is deal with the scenario where superUser checkbox is disabled on the PermissionsMapEditor.
   * If superUser is checked and disabled, this compact editor shouldn't make any changes.
   * However, if superUser isn't checked, this editor should be allowed to make changes.
   * (but it won't be able to set the role as superUser as it's part of the functionality of this editor)
   *
   * @private
   * @returns {Promise.<boolean>} promise to be resolved with true/false
   */
  PermissionsMapGridEditor.prototype.$editorEnabled = function () {
    var userRoles, that = this;

    return that.$getCurrentUser().then(function (currentUser) {
      userRoles = currentUser.get('roles');

      if (!userRoles) { return true; }  // => Grid will be enabled (ie let the server side decide whether any changes are OK)

      return Promise.all(
        userRoles.split(',').map(function (roleName) {
          return that.$getRole(roleName);
        })
      ).then(function (userRoles) {
        for (var i = 0; i < userRoles.length; i++) {
          if (!userRoles[i]) { continue; }

          if (userRoles[i].getType().is('baja:AdminRole')) {
            return true; // The admin role is a super user => grid should be enabled
          }

          if (!userRoles[i].get('enabled')) { continue; }

          var rolePerms = userRoles[i].get('permissions');
          if (rolePerms.isSuperUser()) { return true; }  // user has super user role => grid should be enabled
        }

        return !that.value().isSuperUser();  // Current user doesn't have super user role => grid will only
                                             // be enabled if the current PermissionsMap is not a super user
      });
    })
    .catch(function () {
      // if any of the above fails the default position for the user's permissions is baja.PermissionsMap.SUPER_USER
      return Promise.resolve(true);  // => Grid will be enabled
    });
  };

  /**
   * Get the maximum permissions for this user.
   *
   * @private
   * @returns {Promise.<baja.PermissionsMap>} promise to be resolved with a `baja.PermissionsMap` value
   */

  PermissionsMapGridEditor.prototype.$getMaxPermissions = function () {
    var userRoles, that = this;

    return that.$getCurrentUser().then(function (currentUser) {

      userRoles = currentUser.get('roles');

      if (!userRoles) { return baja.PermissionsMap.SUPER_USER; } // ie let the server side decide whether any changes are OK

      return Promise.all(
        userRoles.split(',').map(function (roleName) {
          return that.$getRole(roleName);
        })
      ).then(function (userRoles) {

        // combine the permissions for all of this user's role to get the max permissions applicable
        var max = baja.PermissionsMap.DEFAULT;

        for (var i = 0; i < userRoles.length; i++) {
          if (!userRoles[i]) { continue; }

          // The admin role is a super user.
          if (userRoles[i].getType().is('baja:AdminRole')) {
            max = baja.PermissionsMap.SUPER_USER;
            break;
          }

          if (!userRoles[i].get('enabled')) { continue; }

          var rolePerms = userRoles[i].get('permissions');
          max = max.or(rolePerms);
        }

        return max;
      });
    })
    .catch(function () {
      // if any of the above fails the default position for the user's permissions is baja.PermissionsMap.SUPER_USER
      // The server side should have the final say on whether any changes are allowed.
      return Promise.resolve(baja.PermissionsMap.SUPER_USER);
    });
  };

  PermissionsMapGridEditor.prototype.$getRoleService = function () {
    return this.getOrdBase().then(getRoleService);
  };

  PermissionsMapGridEditor.prototype.$getUserService = function () {
    return this.getOrdBase().then(getUserService);
  };

  PermissionsMapGridEditor.prototype.$getCategoryService = function () {
    return this.getOrdBase().then(getCategoryService);
  };

  PermissionsMapGridEditor.prototype.$getCurrentUser = function () {
    return this.$getUserService()
      .then(function (userService) {
        var user = userService.get(baja.getUserName());
        return user && user.lease().then(_.constant(user));
      })
      .catch(_.constant(null));
  };

  PermissionsMapGridEditor.prototype.$getRole = function (roleName) {
    return this.$getRoleService()
      .then(function (roleService) {
        var role = roleService.get(roleName);
        return role && role.lease().then(_.constant(role));
      })
      .catch(_.constant(null));
  };

  PermissionsMapGridEditor.prototype.$getCategories = function () {
    return this.$getCategoryService()
      .then(function (categoryService) {
        var categories = categoryService.getSlots().is('baja:Category').toValueArray(),
          i, category, arraySize, categoryNames = [];

        // Add named categories to the categoryNames
        for (i = 0; i < categories.length; i++) {
          category = categories[i];
          categoryNames[category.getIndex()] = category.getDisplayName();
        }

        // Fill in the blanks with the generic name
        arraySize = Math.max(categoryNames.length, MIN_GRID_ROWS+1);
        for (i = 0; i < arraySize; i++) {
          // don't set up index 0 (it's not used for permissions)
          if(i > 0 && !categoryNames[i]) {
            categoryNames[i] = wbutilLex.get('category.generic', [i]);
          }
        }
        return categoryNames;
      })
      .catch(_.constant(null));
  };

  /**
   * Destroys all child editors.
   *
   * @returns {Promise} promise to be resolved when all child editors
   * have been destroyed
   */
  PermissionsMapGridEditor.prototype.doDestroy = function () {
    return this.getChildEditors().destroyAll();
  };

  return (PermissionsMapGridEditor);
});
