/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

define(['baja!',
        'baja!control:PriorityLevel',
        'lex!baja,control',
        'Promise',
        'nmodule/webEditors/rc/fe/baja/util/rangeUtils'], function (
         baja,
         types,
         lexs,
         Promise,
         rangeUtils) {

  'use strict';

  var bajaLex = lexs[0],
      controlLex = lexs[1],
      getEnumRangeDisplay = rangeUtils.getEnumRangeDisplay;

  var lexKeys = [
    'Status.disabled',
    'Status.fault',
    'Status.down',
    'Status.alarm',
    'Status.stale',
    'Status.overridden',
    'Status.null',
    'Status.unackedAlarm'
  ];

  var ACTIVE_LEVEL = 'activeLevel';

  /**
   * @private
   * @exports nmodule/webEditors/rc/fe/baja/util/statusUtils
   */
  var exports = {};


  //TODO: baja.lex does not document module param
  //TODO: BStatus gets a lot of its toString from BBitString - do we need to support that?

  function statusBitsToString(bits) {
    var statuses, i;

    if (bits === 0) {
      statuses = [bajaLex.get('Status.ok')];
    } else {
      statuses = [];

      for (i = 0; i < lexKeys.length; i++) {
        if (bits & (1 << i)) {
          statuses.push(bajaLex.get(lexKeys[i]));
        }
      }
    }

    return '{' + statuses.join(',') + '}';
  }

  /**
   * `DynamicEnum`s don't have display tags, but `FrozenEnum`s do. This function
   * converts a `DynamicEnum` backed by a `PriorityLevel` into the corresponding
   * display tag, pulling it from the `control` lexicon in the case of
   * `fallback`. Other Simple values just get toStringed.
   *
   * @private
   * @inner
   * @param {baja.Simple} activeLevel
   * @returns {String}
   * @see BPriorityLevel.getDisplayTag()
   */
  function activeLevelToString(activeLevel) {
    return '@ ' + (toPriorityLevelString(activeLevel) || activeLevel);
  }

  //TODO: this needs to be a PriorityLevel type extension (NCCB-24257)
  function toPriorityLevelString(priorityLevel) {
    if (!baja.hasType(priorityLevel, 'baja:DynamicEnum')) { return null; }

    var frozen = priorityLevel.getRange().getFrozenType();
    if (frozen && frozen.is('control:PriorityLevel')) {
      var frozenEnum = baja.$('control:PriorityLevel').get(priorityLevel.getOrdinal());
      if (frozenEnum.getTag() === 'fallback') {
        return controlLex.get('def');
      } else {
        return frozenEnum.getDisplayTag();
      }
    }
  }

  function facetsToString(facets) {
    var keys = facets.getKeys(),
        key,
        value,
        a = [],
        i;

    for (i = 0; i < keys.length; i++) {
      key = keys[i];
      value = facets.get(key);

      if (key === ACTIVE_LEVEL) {
        a.push(activeLevelToString(value));
      } else if (value === true) {
        a.push(key);
      } else {
        a.push(key + '=' + value);
      }
    }

    return a.join(' ');
  }

  function simpleToString(simple, obj) {
    switch (String(simple.getType())) {

      case 'baja:Boolean':
        var text = obj[simple ? 'trueText' : 'falseText'];
        return text ? baja.Format.format(text) : bajaLex.get(String(simple));

      case 'baja:Double':
        return simple.toString(obj);

      case 'baja:DynamicEnum':
        return getEnumRangeDisplay(simple.getOrdinal(), obj.range || simple.getRange());

      case 'baja:String':
        return simple;

      default:
        return String(simple);
    }
  }

  /**
   * Converts a `baja.Status` to a display string in a similar way to `BStatus`.
   *
   * @param {baja.Status|Number} status the status to encode (a bit string in
   * the form of a number will be converted to a `baja.Status`)
   * @returns {Promise} promise to be resolved with the toString-ed
   * status, or rejected if the input was invalid
   */
  exports.statusToString = function statusToString(status) {
    if (typeof status === 'number') {
      status = baja.Status.make(status);
    }

    if (!baja.hasType(status, 'baja:Status')) {
      return Promise.reject(new Error('baja:Status or number required'));
    }

    var str = statusBitsToString(status.getBits()),
        facets = status.getFacets();

    if (facets && (facets !== baja.Facets.DEFAULT)) {
      str += ' ' + facetsToString(facets);
    }

    return Promise.resolve(str);
  };


  //TODO: remove 90% of this and use baja.Simple#toString(cx)
  /**
   * Resolve a human-readable string value for a `StatusValue`.
   *
   * Supports `trueText` and `falseText` for booleans, `range` for enums,
   * and all properties (`precision` etc.) that are supported by `numberUtils`.
   *
   * @param {baja.Struct} statusValue a `baja:StatusValue`
   * @param {Object} [config] properties to use to format the `value` property
   * @returns {*}
   */
  exports.statusValueToString = function (statusValue, config) {
    if (!baja.hasType(statusValue, 'baja:StatusValue')) {
      return Promise.reject(new Error('baja:StatusValue required'));
    }

    var status = statusValue.get('status'),
        value = statusValue.get('value');

    return Promise.join(
      status.isNull() ? '-' : simpleToString(value, config || {}),
      exports.statusToString(status)
    )
      .spread(function (valueString, statusString) {
        return valueString + ' ' + statusString;
      });
  };

  return (exports);
});
