/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/util/typeUtils
 */
define(['baja!',
        'lex!',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/util/textUtils'], function (
        baja,
        lexjs,
        Promise,
        _,
        textUtils) {

  'use strict';

  var knownBadTypes = {},
      toFriendlyString = textUtils.toFriendly;

  /**
   * Is this string a valid baja type spec?
   *
   * @inner
   * @type {RegExp}
   */
  var TYPE_SPEC_REGEX = /^[A-Za-z][A-Za-z0-9_]*:[A-Z][A-Za-z0-9_]*$/;

  /**
   * Search through the type for its contract info for the given slot.
   *
   * @inner
   * @param {Type} type
   * @param {String|baja.Slot} slot
   * @returns {Object}
   * @throws {Error} if the slot could not be found on the type
   */
  function findInContract(type, slot) {
    if (!slot) {
      throw new Error('slot required');
    }

    var inType = type,
        slotName = String(slot),
        contract, obj, i;

    while (type && (contract = type.getContract())) {
      for (i = 0; i < contract.length; i++) {
        obj = contract[i];
        if (obj.n === slotName) {
          return obj;
        }
      }

      type = type.getSuperType();
    }

    throw new Error('slot ' + slotName + ' not found on ' + inType + ' or any supertype');
  }

  /**
   * @private
   * @exports nmodule/webEditors/rc/fe/baja/util/typeUtils
   */
  var exports = {};

  //TODO: BajaScript should eventually give a nice function for this.
  /**
   * Return the slot display name for a certain slot on a type.
   *
   * @param {Type} type
   * @param {String|baja.Slot} slotName
   * @returns {String} slot display name
   * @throws {Error} if slot was not found on the Type
   */
  exports.getSlotDisplayName = function (type, slotName) {
    return findInContract(type, slotName).dn;
  };

  //TODO: BajaScript should eventually give a nice function for this.
  /**
   * Return the slot facets for a certain slot on a type.
   *
   * @param {Type} type
   * @param {String|baja.Slot} slotName
   * @returns {baja.Facets}
   * @throws {Error} if slot was not found on the Type
   */
  exports.getSlotFacets = function (type, slotName) {
    var obj = findInContract(type, String(slotName)),
        facets = baja.Facets.DEFAULT;

    return obj.x ? facets.decodeFromString(obj.x) : facets;
  };

  /**
   * Return the Type of the given slot. If the slot is an action or topic,
   * return null.
   *
   * @param {Type} type
   * @param {String|baja.Slot} slotName
   * @returns {Type|null} the Type of the given slot, or `null` if it's not a
   * Property
   * @throws {Error} if slot was not found on the Type
   */
  exports.getSlotType = function (type, slotName) {
    var obj = findInContract(type, String(slotName));

    if (obj.v) {
      return baja.lt(obj.v.t);
    }

    return null;
  };

  /**
   * Flatten out a type's supertype chain into an array of Types, `baja:Object`
   * first.
   *
   * @param {Type} type
   * @returns {Array.<Type>}
   */
  exports.getSuperTypeChain = function (type) {
    var arr = [];
    while (type) {
      arr.push(type);
      type = type.getSuperType();
    }
    return arr.reverse();
  };

  /**
   * Get a unique slot name for adding a new child component to a parent.
   *
   * @param {baja.Component} comp a component to which we're adding a slot
   * @param {String|baja.Simple|Type} [typeSpec] if given, the unique name will
   * match the type name
   * @returns {String} a unique name to use to add a new slot
   */
  exports.getUniqueSlotName = function (comp, typeSpec) {
    if (!typeSpec) {
      return comp.getUniqueName('newSlot');
    }

    var type = baja.lt(String(typeSpec));

    if (!type) {
      throw new Error('type ' + typeSpec + ' not found or not imported');
    }

    return comp.getUniqueName(type.getTypeName());
  };

  /**
   * Convenience method wrapping `baja.importTypes` in a promise.
   *
   * @param {Array.<String>} [types] type array. Type specs can also be passed
   * as direct parameters.
   *
   * @returns {Promise}
   *
   * @example
   *   <caption>Import an array of type specs</caption>
   *
   *   var specs = [ 'baja:Double', 'baja:String', 'baja:Time' ];
   *   typeUtils.importTypes(specs)
   *     .then(function (types) {
   *       //array of types
   *     });
   *
   * @example
   *   <caption>Import type specs as parameters</caption>
   *
   *   typeUtils.importTypes('baja:Double', 'baja:String', 'baja:Time')
   *     .spread(function (doubleType, stringType, timeType) {
   *       //types as parameters
   *     });
   */
  exports.importTypes = function (types) {
    var isArray = Array.isArray(types),
        arr = isArray ? types : Array.prototype.slice.call(arguments);

    return Promise.try(function () {
      _.each(arr, function (typeSpec) {
        if (knownBadTypes[typeSpec]) {
          //if we know we've got a bad Type, save ourselves the network call
          throw new Error('unknown type: ' + typeSpec);
        }
      });

      if (!arr[0]) { //nothing to resolve
        return [];
      } else {
        return baja.importTypes({ typeSpecs: arr })
          .catch(function (err) {
            _.each(err, function (val, typeSpec) {
              knownBadTypes[typeSpec] = true;
            });
            throw err;
          });
      }
    });
  };

  /**
   * @param {*} obj
   * @returns {Boolean} true if the object is an instance of `baja:Value`.
   */
  exports.isValue = function (obj) {
    return baja.hasType(obj, 'baja:Value');
  };

  /**
   * @param {*} obj
   * @returns {Boolean} true if the object is an instance of `baja:Simple`.
   */
  exports.isSimple = function (obj) {
    return baja.hasType(obj, 'baja:Simple');
  };

  /**
   * Return true if the object is an instance of `baja:Number`. Note that all
   * of these implement `valueOf()` and so can be used in mathematical
   * calculations.
   *
   * @param {*} obj
   * @returns {Boolean}
   */
  exports.isNumber = function (obj) {
    return baja.hasType(obj, 'baja:Number');
  };

  /**
   * @param {Object} obj
   * @returns {Boolean} true if the object is an instance of `baja:Complex`.
   */
  exports.isComplex = function (obj) {
    return baja.hasType(obj, 'baja:Complex');
  };

  /**
   * @param {*} obj
   * @returns {Boolean} true if the object is an instance of `baja:Component`.
   */
  exports.isComponent = function (obj) {
    return baja.hasType(obj, 'baja:Component');
  };

  /**
   * Return true if the parameter is a type, or type spec.
   *
   * @param {Object|String} type
   * @returns {Boolean} true if the object is an instance of a baja type, or
   * a string that looks like a baja type spec
   */
  exports.isType = function (type) {
    if (type) {
      return TYPE_SPEC_REGEX.test(type) ||
        typeof type.getTypeSpec === 'function';
    }
    return false;
  };

  /**
   * Check to see if one constructor is a subclass of another constructor.
   *
   * @param {Function|Object} superClass constructor function, or object whose
   * constructor to check
   * @param {Function|Object} subClass constructor function, or object whose
   * constructor to check
   * @returns {Boolean} true if subClass inherits from superClass, or if they
   * are the same constructor
   */
  exports.isAssignableFrom = function (superClass, subClass) {
    if (!superClass || !subClass) {
      return false;
    }

    var superCtor = typeof superClass === 'function' ?
          superClass :
          superClass.constructor,
        subCtor = typeof subClass === 'function' ?
          subClass :
          subClass.constructor;

    return Object.create(subCtor.prototype) instanceof superCtor;
  };

  /**
   * Creates a `baja:TypeSpec` instance from the given type spec string.
   *
   * @param {baja.Value|String|Type} str type spec (to be toStringed)
   * @returns {baja.Value} a `baja:TypeSpec` instance, suitable for adding to
   * a Component
   */
  exports.toTypeSpec = function (str) {
    return baja.$('baja:TypeSpec', String(str));
  };

  /**
   * Returns a Promise that will resolve to a suitable display
   * name string for the given type. This will try to obtain a string
   * from the lexicon for the module or, if no entry is found, it
   * will return a 'friendly' string from the name contained in
   * the type spec.
   *
   * @param {Type|String} typeSpec
   * @returns {Promise.<String>}
   */
  exports.getTypeDisplayName = function (typeSpec) {
    var type = baja.lt(String(typeSpec)),
        name = type.getTypeName(),
        module = type.getModuleName();

    return lexjs.module(module)
      .then(function (lex) {
        return lex.get(name + '.displayName') || toFriendlyString(name);
      });
  };

  return (exports);
});
