/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * @module nmodule/webChart/rc/model/BaseSeries
 */
define([
  'd3',
  'jquery',
  'Promise',
  'baja!',
  'nmodule/webChart/rc/choiceUtil',
  'nmodule/webChart/rc/export/exportUtil',
  'nmodule/webChart/rc/webChartUtil',
  'nmodule/webChart/rc/model/modelUtil',
  'nmodule/webChart/rc/model/samplingUtil',
  'nmodule/webChart/rc/model/ValueScale',
  'nmodule/webChart/rc/chartEvents',
  'nmodule/webEditors/rc/fe/baja/util/numberUtils' ], function (
   d3,
   $,
   Promise,
   baja,
   choiceUtil,
   exportUtil,
   webChartUtil,
   modelUtil,
   samplingUtil,
   ValueScale,
   events,
   numberUtils) {

  "use strict";

  /**
   * BaseSeries represents the Base Series for a chart data representation.
   *
   * @class
   * @alias module:nmodule/webChart/rc/model/BaseSeries
   * @param {module:nmodule/webChart/rc/model/BaseModel} model The model used with this Series.
   * @param {Object} params The container for the parameters for creating this Series.
   * @param {String} params.ord The ord for the Series.
   * @param {String} params.originalOrd The original ord for the Series (prior
   * to traversing to history space via a history extension or n:history tag)
   * @param {baja.Value} [params.value] Optional resolved value for the Series.
   */
  var BaseSeries = function (model, params) {
    var that = this;
    if (!params) {
      params = {};
    }
    that.$ord = webChartUtil.getRelativeOrd(params.ord);
    that.$originalOrd = webChartUtil.getRelativeOrd(params.originalOrd) || that.$ord;
    that.$value = params.value || null;
    that.$model = model || null;
    that.$displayName = null;
    that.$displayPath = null;
    that.$shortDisplayName = null;
    that.$units = baja.Unit.DEFAULT;
    that.$points = [];
    that.$enabled = true;
    that.$subscribeContinuous = false;
    that.$capacity = webChartUtil.getMaxSeriesCapacity();
    that.$lastY = null;
    that.$facets = baja.Facets.DEFAULT;
    that.$loaded = false;
    that.$loading = false;
    that.$recordType = null;
    that.$valueScale = new ValueScale(model, that); //fallback scale
    that.$cachedPrecision = null;

    that.$focusPoints = null; //subset of points in focus, no sampling yet
    that.$samplingPoints = null; //subset of points in focus, then sample algorithm applied

    that.$error = null; //for user error reporting

    //initialize settings based on params
    if (params && model) {
      model.settings().seriesSettings(that, params);
    }
  };

  /**
   * Color accessor. Get or set Series color as an HTML string '#rrggbb'
   *
   * @param {String} [color] If specified, this will set color.
   * @returns {String}
   */
  BaseSeries.prototype.color = function (color) {
    return this.$model.settings().seriesColor(this, color);
  };

  /**
   * ChartType accessor.
   *
   * @param {String} [chartType] If specified, this will set chartType.
   * @returns {String}
   */
  BaseSeries.prototype.chartType = function (chartType) {
    return this.$model.settings().seriesChartType(this, chartType);
  };

  /**
   * Return true when the Series is considered a Line Chart.
   * @returns {boolean}
   */
  BaseSeries.prototype.isLine = function () {
    var chartType = this.$model.settings().seriesChartType(this);
    return chartType === "line" || chartType === "discreteLine";
  };

  /**
   * Return true when the Series is considered a Discrete Line.
   * @returns {boolean}
   */
  BaseSeries.prototype.isDiscreteLine = function () {
    var chartType = this.$model.settings().seriesChartType(this);
    return chartType === "discreteLine";
  };

  /**
   * Return true when the Series is considered a BarChart.
   * @returns {boolean}
   */
  BaseSeries.prototype.isBar = function () {
    var chartType = this.$model.settings().seriesChartType(this);
    return chartType === "bar";
  };

  /**
   * Return true when the Series is considered a Shade Chart.
   * @returns {boolean}
   */
  BaseSeries.prototype.isShade = function () {
    var chartType = this.$model.settings().seriesChartType(this);
    return chartType === "shade";
  };

  /**
   * History capacity accessor. Note that you should not set the capacity greater than the Max Series Capacity that defaults
   * to 250K records.
   *
   * @param {Number} [capacity] If specified, this will set capacity.
   * @returns {Number}
   */
  BaseSeries.prototype.capacity = function (capacity) {
    var that = this;
    if (capacity !== undefined) {
      //only change capacity if its less than max and greater than zero
      if (capacity > 0 && capacity < webChartUtil.getMaxSeriesCapacity()) {
        that.$capacity = capacity;
      }
    }
    return that.$capacity;
  };

  /**
   * Trim the Points the allotted capacity.
   */
  BaseSeries.prototype.trimToCapacity = function () {
    var that = this,
        cap    = that.$capacity,
        length = that.$points ? that.$points.length : -1;

    if (cap > -1 && length > cap) {
      that.$points = that.$points.slice(length - cap);
      if (!that.$model.isOverMaxCapacity() && length > webChartUtil.getMaxSeriesCapacity()) {
        that.$model.setOverMaxCapacity(true);
      }
    }
  };

  /**
   * Error accessor.
   *
   * @param {String} [error] If specified, this will set the error for the series.
   * @returns {String}
   */
  BaseSeries.prototype.error = function (error) {
    if (error !== undefined) {
      this.$error = error;
    }
    return this.$error;
  };

  /**
   * Get the facets for this series.
   *
   * @returns {baja.Facets}
   */
  BaseSeries.prototype.facets = function () {
    return this.$facets;
  };

  /**
   * Return true if concurrent load is allowed, otherwise load is complete after others are loaded
   *
   * @returns {Boolean}
   */
  BaseSeries.prototype.isPrimaryLoad = function () {
    return true;
  };

  /**
   * This type describes the Object properties expected for a Point used by the Chart Model.
   *
   * @typedef {Object} module:nmodule/webChart/rc/model/BaseSeries~Point
   * @property {Data|Object} x The value for the x axis, usually a Date.
   * @property {Number|Object} y The value for the y axis, usually a Number.
   * @property {boolean} [skip] if given, this tells the ChartModel that there is break in the data.
   * @property {Number} [status] if given, this is the 'status bits' of a baja.Status.

   /**
   * Data Points accessor.
   *
   * @param {Array.<module:nmodule/webChart/rc/model/BaseSeries~Point>} [points] If specified, this will set the data points Array for the series.
   * @returns {Array.<module:nmodule/webChart/rc/model/BaseSeries~Point>}
   */
  BaseSeries.prototype.points = function (points) {
    var that  = this;
    if (points !== undefined) {
      that.$points = points;
      that.trimToCapacity();
    }
    return that.$points;
  };

  /**
   * Focus Points accessor returns a subset of the total points if user is zoomed in on a subset of domain.
   * @param {boolean} store If true, store the results for other calls to focusPoints that are not for storage.
   *                  If false, return the existing stored focusPoints if the points are already available.
   * @returns {Array.<module:nmodule/webChart/rc/model/BaseSeries~Point>}
   */
  BaseSeries.prototype.focusPoints = function (store) {

    var that = this;
    if (that.$focusPoints && !store) {
      return that.$focusPoints;
    }

    var points   = that.points(),
        model    = that.$model,
        scale    = model.timeScale().scale(),
        domain   = scale.domain(),
        getX     = function (point) {
          return point.x;
        },
        index    = samplingUtil.getEdgeIndex(points, domain[0], getX),
        endIndex = samplingUtil.getEdgeIndex(points, domain[1], getX);


    endIndex += 2; //always go at least 2

    var zoomedPoints = points.slice(index, endIndex);
    if (store) {
      that.$focusPoints = zoomedPoints;
    }
    return zoomedPoints;
  };

  /**
   * Set the samplingPoints for this series
   * @param {Array.<module:nmodule/webChart/rc/model/BaseSeries~Point>} samplingPoints
   */
  BaseSeries.prototype.setSamplingPoints = function (samplingPoints) {
    this.$samplingPoints = samplingPoints;
  };

  /**
   * Sampling Points accessor returns a subset of the total points if user is zoomed in on a subset of domain
   *
   * @returns {Array.<module:nmodule/webChart/rc/model/BaseSeries~Point>}
   */
  BaseSeries.prototype.samplingPoints = function () {

    var that     = this,
        sampling = that.$model.isSampling();

    if ((sampling || that.isBar()) && that.$samplingPoints) {
      return that.$samplingPoints;
    }

    return that.focusPoints();
  };

  /**
   * Units accessor.
   *
   * @param {baja.Unit} [units] If specified, this will set the units.
   * @returns {baja.Unit}
   */
  BaseSeries.prototype.units = function (units) {
    if (units !== undefined) {
      this.$units = units;
    }
    return this.$units;
  };

  /**
   * Value accessor.
   *
   * @param {baja.Value} [value] If specified, this will set the value.
   * @returns {baja.Value}
   */
  BaseSeries.prototype.value = function (value) {
    if (value !== undefined) {
      this.$value = value;
    }
    return this.$value;
  };

  /**
   * ValueScale accessor.
   *
   * @returns {module:nmodule/webChart/rc/model/ValueScale}
   */
  BaseSeries.prototype.valueScale = function () {
    return this.$valueScale;
  };

  /**
   * Display Name accessor.
   *
   * @param {String} [displayName] If specified, this will set the displayName.
   * @returns {String}
   */
  BaseSeries.prototype.displayName = function (displayName) {
    if (displayName !== undefined) {
      this.$displayName = displayName;
    }

    return this.$displayName;
  };

  /**
   * Display Path accessor.
   * @param {String} [displayPath] If specified, this will set the displayPath.
   *
   * @returns {String}
   */
  BaseSeries.prototype.displayPath = function (displayPath) {
    if (displayPath !== undefined) {
      this.$displayPath = displayPath;
    }

    if (!this.$displayPath) {
      return this.$ord;
    }

    return this.$displayPath;
  };

  /**
   * Short Display Name accessor.
   * @param {String} [shortDisplayName] If specified, this will set the shortDisplayName.
   *
   * @returns {String}
   */
  BaseSeries.prototype.shortDisplayName = function (shortDisplayName) {
    if (shortDisplayName !== undefined) {
      this.$shortDisplayName = shortDisplayName;
    }

    if (!this.$shortDisplayName) {
      return webChartUtil.getShortDisplayName(this.displayName());
    }

    return this.$shortDisplayName;
  };

  /**
   * Is this Series enabled.
   *
   * @returns {boolean}
   */
  BaseSeries.prototype.isEnabled = function () {
    return this.$enabled;
  };
  /**
   * set enabled.
   *
   * @param {boolean} enabled This will set the enabled attribute of the series.
   */
  BaseSeries.prototype.setEnabled = function (enabled) {
    var that = this;
    if (enabled !== that.$enabled) {
      that.$enabled = enabled;

      if (that.$model) {
        that.$model.jq()
          .trigger(enabled ? events.SERIES_ENABLED : events.SERIES_DISABLED, [ that ]);
      }
    }
  };
  /**
   * ord accessor.
   *
   * @returns {baja.Ord}
   */
  BaseSeries.prototype.ord = function () {
    return this.$ord;
  };

  /**
   * Original Ord accessor.
   *
   * @returns {baja.Ord}
   */
  BaseSeries.prototype.originalOrd = function () {
    return this.$originalOrd;
  };

  /**
   * Get the recordType.
   *
   * @returns {baja.TypeSpec}
   */
  BaseSeries.prototype.recordType = function () {
    return this.$recordType;
  };

  /**
   * loaded accessor.
   *
   * @returns {boolean}
   */
  BaseSeries.prototype.isLoaded = function () {
    return this.$loaded;
  };

  /**
   * Override this method to load your series info and resolve a Promise when complete.
   *
   * @returns {Promise}
   */
  BaseSeries.prototype.loadInfo = function () {
    this.$loaded = true;
    return modelUtil.resolvedPromise();
  };

  /**
   * Override this method to load your series data and resolve a promise when complete.
   *
   * @returns {Promise}
   */
  BaseSeries.prototype.loadData = function () {
    this.$loaded = true;
    return modelUtil.resolvedPromise();
  };

  /**
   * A Series is Discrete if its record type is a Boolean or DynamicEnum.
   * @returns {boolean}
   */
  BaseSeries.prototype.isDiscrete = function () {
    var recordType = this.$recordType;
    return recordType === "baja:Boolean" || recordType === "baja:DynamicEnum";
  };

  /**
   * A Series is Boolean if its record type is a Boolean.
   * @returns {boolean}
   */
  BaseSeries.prototype.isBoolean = function () {
    var recordType = this.$recordType;
    return recordType === "baja:Boolean";
  };

  /**
   *What is the default d3 Interpolation for the line? Some options are the following:
   *
   * linear - piecewise linear segments, as in a polyline.
   *
   * step-after - alternate between horizontal and vertical segments, as in a step function.
   *
   * step - alternate between horizontal and vertical segments, as in a step function.
   *
   * @returns {String}
   */
  BaseSeries.prototype.getDefaultInterpolation = function () {
    return "linear";
  };

  /**
   * What is the current Interpolation for the line?
   * @returns {String}
   */
  BaseSeries.prototype.getLineInterpolation = function () {
    if (this.isDiscreteLine()) {
      return "step-after";
    } else {
      return "linear";
      //return "monotone"; //TODO: although this looks much better, not safe due to NCCB-7988
    }
  };

  /**
   * Get the Display Value for y. Override for custom values in data.
   * @param y the raw y data
   * @returns {Promise.<String>}
   */
  BaseSeries.prototype.resolveYDisplay = function (y) {
    var that   = this,
        facets = that.$facets;

    if (that.isDiscrete()) {
      return modelUtil.resolveEnumDisplay(that, y, facets);
    }

    return modelUtil.resolveNumericDisplay(y, that.units(), that.getPrecision());
  };

  /**
   * Return the precision for this Series based on the available facets
   * @returns {Number} min 0, max 20
   */
  BaseSeries.prototype.getPrecision = function () {
    var that   = this,
        facets = that.$facets,
        cached = that.$cachedPrecision,
        precision;

    if (cached !== null) {
      return cached;
    }

    if (!facets) {
      return 2; //default for number until facets is available
    }
    //TODO: why do we allow entering a negative precision?
    precision = Math.max(0, Math.min(facets.get("precision", 2), 20));
    that.$cachedPrecision = precision;
    return precision;
  };

  /**
   * Provide either the number of ticks desired, hardcoded tick values, of function
   * to pass to d3.scale.ticks.
   * @returns {number|Array.<number>|*}
   */
  BaseSeries.prototype.getTicks = function () {

    var that = this,
      newTicks = [];

    if (that.isShade()) {
      return 0;
    } else if (that.isBoolean()) {
      var domain = that.valueScale().scale().domain(),
        min = domain[0],
        max = domain[1];

      if (min <= 0 && max >= 0) {
        newTicks.push(0);
      }

      if (min <= 1 && max >= 1) {
        newTicks.push(1);
      }
      return newTicks;
    } else if (that.isDiscrete()) {
      var ticks = that.valueScale().scale().ticks(8);

      for (var i = 0; i < ticks.length; i++) {
        var tick = Math.round(ticks[i]);
        if (!newTicks.contains(tick)) {
          newTicks.push(tick);
        }
      }

      return newTicks;

    }
    return 8; //TODO: maybe a bajaux property?
  };

  /**
   * Return true if the subscription since the last update is continuous
   *
   * @returns {boolean}
   */
  BaseSeries.prototype.isSubscribeContinuous = function () {
    return this.$subscribeContinuous;
  };

  /**
   * Set Subscribe Continuous.
   *
   * @param {boolean} [subscribeContinuous] If specified, this will set subscribeContinuous.
   */
  BaseSeries.prototype.setSubscribeContinuous = function (subscribeContinuous) {
    if (subscribeContinuous !== undefined) {
      this.$subscribeContinuous = subscribeContinuous;
    }
  };

  /**
   * Unload the current series by emptying the points and set loaded status back to false.
   */
  BaseSeries.prototype.unload = function () {
    var that = this;

    that.$points.length = 0;
    that.$loaded = false;
    that.$loading = false;
  };

  /**
   * Using the raw data, convert to a {@link module:nmodule/webChart/rc/model/BaseSeries~Point|BaseSeries~Point} and add to the
   * BaseSeries.points array
   * @param {Object} raw The object passed over from the WebChartQueryServlet
   */
  BaseSeries.prototype.preparePoint = function (raw) {
  };

  /**
   * Using raw data from a subscription, convert this to a  {@link module:nmodule/webChart/rc/model/BaseSeries~Point|BaseSeries~Point} and
   * add to the BaseSeries.points array. If this is the same as BaseSeries.preparePoint then its doesn't need to be
   * overridden.
   * @param {Object} raw The object passed from subscription
   */
  BaseSeries.prototype.prepareLivePoint = function (raw) {
    this.preparePoint(raw);
  };

  /**
   * Subscribe to the series.
   *
   * @param {baja.Subscriber} [subscriber]
   */
  BaseSeries.prototype.subscribe = function (subscriber) {
  };

  /**
   * Unsubscribe to the series.
   *
   * @param {baja.Subscriber} [subscriber]
   */
  BaseSeries.prototype.unsubscribe = function (subscriber) {
  };

  /**
   * Save the Series to the JSON.
   *
   * @param  {Object} seriesJson The object used to hold the details
   * of the Series data.
   * @param {baja.Component} [optionInfo] has slots ordType and optionally
   *  baseOrd.
   */
  BaseSeries.prototype.saveToJson = function (seriesJson, optionInfo) {
    var that      = this,
        color     = that.color(),
        chartType = that.chartType();

    if (optionInfo) {
      var ordType = choiceUtil.getChoice(optionInfo, "ordType");
      if (ordType === "absolute") {
        seriesJson.ord = that.ord();
      } else {
        var baseOrd = baja.Ord.make(optionInfo.get('baseOrd'));
        seriesJson.ord = exportUtil.relativizeOrd(baseOrd, baja.Ord.make(that.originalOrd())).toString();
      }
    } else {
      seriesJson.ord = that.ord();
    }

    if (chartType) {
      seriesJson.chartType = baja.bson.encodeValue(chartType);
    }
    if (color) {
      seriesJson.color = baja.bson.encodeValue(color);
    }
  };

  return BaseSeries;
});
