/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author JJ Frankovich
 */

/**
 * @module nmodule/webChart/rc/model/TimeScale
 * @private
 */
define([
  'd3',
  'jquery',
  'baja!',
  'moment',
  'nmodule/webChart/rc/webChartUtil',
  'nmodule/webChart/rc/localeUtil',
  'nmodule/webChart/rc/model/modelUtil',
  'nmodule/webChart/rc/model/samplingUtil',
  'nmodule/webChart/rc/model/BaseScale'
], function(d3, $, baja, moment, webChartUtil, localeUtil, modelUtil, samplingUtil, BaseScale) {
  "use strict";

  /**
   * TimeScale represents the time scale for a chart data.
   *
   * @private
   * @class
   * @alias module:nmodule/webChart/rc/model/TimeScale
   * @param {Object} model The model used with this Scale.
   * @param {BaseSeries} [series] The first Series in this scale (if present).
   */
  var TimeScale = function(model) {
    BaseScale.apply(this, arguments);
    var that = this;
    that.$units = baja.Unit.DEFAULT;
    that.$dataProperty = "x";
    that.$scale = d3.time.scale();
    //default d3 multi-scale except that months use abbreviations
    that.$scale.tickFormat = localeUtil.getLocaleTickFormat;

    that.$seriesList = model.seriesList();
  };

  TimeScale.prototype = Object.create(BaseScale.prototype);
  TimeScale.prototype.constructor = TimeScale;

  TimeScale.prototype.requiresTicks = function() {
    return true; //time always requires ticks
  };

  /**
   * SeriesList Accessor
   *
   * @returns {Array.<BaseSeries>}
   */
  TimeScale.prototype.seriesList = function() {
    return this.$seriesList;
  };

  /**
   * Primary Series Accessor
   *
   * @returns {webChart.BaseSeries}
   */
  TimeScale.prototype.primarySeries = function() {
    return this.$seriesList[0];
  };


  /**
   * Scale accessor.
   * @returns {d3.Scale}
   */
  TimeScale.prototype.scale = function() {
    return this.$scale;
  };

  /**
   * Get Time Display and ensure day is printed
   * @param {Date} y
   * @returns {string}
   */
  TimeScale.prototype.getFullTimeDisplay = function(y) {
    return this.getTimeDisplay(y, true);
  };

  /**
   * Add day if needed
   * @param {String} day
   * @param {String} time
   * @param {boolean} ensureDayOn
   * @returns {*}
   */
  function ensureDay (day, time, ensureDayOn) {
    if(ensureDayOn){
      return  day + ", " + time;
    }
    return  time;
  }


  /**
   * In the context of the domain, provide an accurate time.
   * @param {Date} y
   * @param {boolean} [ensureDayOn]
   * @param {boolean} [estimateOk]
   * @returns {String}
   */
  TimeScale.prototype.getTimeDisplay = function(y, ensureDayOn, estimateOk) {
    var that = this,
        format = ensureDayOn ? localeUtil.getD3LongDateFormat() : localeUtil.getD3DateFormat(), //full month text (use more space) or just mm/dd/yyy
        xdomain = that.$scale.domain(),
        start = moment(xdomain[0]),
        end = moment(xdomain[1]),
        startDisplay = localeUtil.format(format, start),
        endDisplay = localeUtil.format(format, end),
        value = moment(y),
        valueDisplay = localeUtil.format(format, value),
        model = that.$model,
        samplingOn = model.isSampling() || modelUtil.hasBar(model.seriesList()),
        samplingMillis = samplingOn ? model.samplingMillis() : "",
        startValue,
        endValue,
        startValueDisplay,
        endValueDisplay;


    if(startDisplay === "Invalid date" || endDisplay === "Invalid date"){
      return "";
    }

    if(samplingMillis){
      startValue = moment(value);
      endValue = modelUtil.getEndTime(model, startValue);

      startValueDisplay = localeUtil.format(format, startValue);
      endValueDisplay = localeUtil.format(format, endValue);
    }

    //same day?
    if(startDisplay === endDisplay){ //is time required?

      if(samplingMillis){
        startValueDisplay = localeUtil.format(localeUtil.getD3HourMinutePeriodsFormat(), startValue);
        endValueDisplay = localeUtil.format(localeUtil.getD3HourMinutePeriodsFormat(), endValue);
      }

      //are seconds required?
      if(moment(value).startOf('minute').isSame(value)){

        if(startValueDisplay !== endValueDisplay){
          return ensureDay(valueDisplay, webChartUtil.lex.get("dateTo", startValueDisplay, endValueDisplay), ensureDayOn);
        }

        return ensureDay(valueDisplay, localeUtil.format(localeUtil.getD3HourMinutePeriodsFormat(), value), ensureDayOn);
      }
      else{

        if(startValueDisplay !== endValueDisplay){ //same minute rollup doesn't need end time
          if(modelUtil.safeMomentAdd(startValue, 60000).isSame(endValue)){
            return ensureDay(valueDisplay, startValueDisplay, ensureDayOn);
          }

          return ensureDay(valueDisplay, webChartUtil.lex.get("dateTo", startValueDisplay, endValueDisplay), ensureDayOn);
        }

        return ensureDay(valueDisplay, localeUtil.format(localeUtil.getD3TimeFormat(), value), ensureDayOn);
      }
    }
    else if(moment(value).startOf('day').isSame(value)){ //everything is same day
      if(samplingMillis){
        if(startValueDisplay !== endValueDisplay){
          return webChartUtil.lex.get("dateTo", startValueDisplay, endValueDisplay);
        }
      }

      return valueDisplay;
    }
    else{

      if(estimateOk){ //not the same day. if estimate ok, then no seconds required
        return valueDisplay + " " + localeUtil.format(localeUtil.getD3HourMinutePeriodsFormat(), value);
      }
      return valueDisplay + " " +  localeUtil.format(localeUtil.getD3TimeFormat(), value);
    }
  };

  return TimeScale;
});
