/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/*jshint forin: false */
/*global localStorage, sessionStorage */

/**
 * @private
 * @module nmodule/webEditors/rc/util/storageUtil
 */
define([], function () {
  'use strict';

  var _local, 
    _session, 
    _localStorage;

  try {
    _localStorage = localStorage;
  } 
  catch(ignore) {} 

  /**
   * @inner
   * @param {Object} storage
   * @returns {boolean}
   */
  function storageWorks(storage) {
    var x = '__webEditors';
    try {
      storage.setItem(x, x);
      storage.removeItem(x);
      return true;
    } catch (e) { return false; }
  }

  /**
   * @private
   * @exports nmodule/webEditors/rc/util/storageUtil
   */
  var exports = {};

  /**
   * Return true if a valid `localStorage` exists.
   *
   * @returns {boolean}
   */
  exports.hasLocalStorage = function () {
    return storageWorks(_localStorage);
  };

  /**
   * Return the local storage object to use. If a valid `localStorage` object
   * cannot be found, one will be shimmed in (note that it will be cleared
   * when the page is closed or refreshed).
   *
   * @returns {Object}
   */
  exports.getLocalStorage = function () {
    if (!_local) {
      _local = exports.hasLocalStorage() ?
        _localStorage :
        exports.$polyfill();
    }
    return _local;
  };

  /**
   * Return true if a valid `sessionStorage` exists.
   *
   * @returns {boolean}
   */
  exports.hasSessionStorage = function () {
    return storageWorks(sessionStorage);
  };

  /**
   * Return the session storage object to use. If a valid `sessionStorage`
   * object cannot be found, one will be shimmed in (note that it will be
   * cleared when the page is closed or refreshed).
   *
   * @returns {Object}
   */
  exports.getSessionStorage = function () {
    if (!_session) {
      _session = exports.hasSessionStorage() ?
        sessionStorage :
        exports.$polyfill();
    }

    return _session;
  };

  //TODO: this should really be some OSS shim but probably too late to get approved at this point
  //http://dev.w3.org/html5/webstorage/#storage-0
  /**
   * Create a new Storage polyfill object to use when a valid
   * `localStorage`/`sessionStorage` object cannot be found.
   *
   * @private
   * @returns {Object}
   */
  exports.$polyfill = function () {
    var obj = {};

    Object.defineProperty(obj, 'length', {
      enumerable: false,
      get: function () {
        var length = 0, prop;
        for (prop in this) {
          length++;
        }
        return length;
      }
    });

    Object.defineProperty(obj, 'getItem', {
      enumerable: false,
      value: function (key) {
        if (arguments.length === 0) {
          throw new Error('1 argument required');
        }

        if (this.hasOwnProperty(String(key))) {
          return this[key];
        }
        return null;
      }
    });

    Object.defineProperty(obj, 'setItem', {
      enumerable: false,
      value: function (key, value) {
        if (arguments.length < 2) {
          throw new Error('2 arguments required');
        }

        this[key] = String(value);
      }
    });

    Object.defineProperty(obj, 'removeItem', {
      enumerable: false,
      value: function (key) {
        if (arguments.length === 0) {
          throw new Error('1 argument required');
        }

        delete this[key];
      }
    });

    Object.defineProperty(obj, 'clear', {
      enumerable: false,
      value: function () {
        for (var prop in this) {
          delete this[prop];
        }
      }
    });

    Object.defineProperty(obj, 'key', {
      enumerable: false,
      value: function (idx) {
        if (arguments.length === 0) {
          throw new Error('1 argument required');
        }

        var i = 0;
        for (var prop in this) {
          if (i === idx) {
            return prop;
          }
          i++;
        }
        return null;
      }
    });

    return obj;
  };

  return exports;
});
