/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/commands/transfer/TransferCommand
 */
define(['bajaux/commands/Command',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/wb/mixin/mixinUtils'], function (
         Command,
         Promise,
         _,
         mixinUtils) {

  'use strict';

  /**
   * Command to execute a transfer operation.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/commands/transfer/TransferCommand
   * @extends {module:bajaux/commands/Command}
   */
  var TransferCommand = function TransferCommand(owner, transferOp, params) {
    this.$op = transferOp;

    var subject = params.subject,
        enabled = owner.isTransferOpEnabled(transferOp, subject),
        supported = owner.isTransferOpSupported(transferOp, subject);

    Command.call(this, {
      displayName: params.displayName || transferOp,
      module: params.module,
      lex: params.lex,
      enabled: enabled && supported,
      func: function () {
        if (!supported) {
          throw new Error(transferOp + ' op not supported');
        }
        return owner.doTransferOp(transferOp, subject);
      }
    });
  };
  TransferCommand.prototype = Object.create(Command.prototype);
  TransferCommand.prototype.constructor = TransferCommand;

  /**
   * Create an instance of `TransferCommand`. The resolved instance, when
   * invoked, will perform the transfer operation on the owner's subject.
   *
   * @param {*} owner - must have `TransferSupport` mixin
   * @param {string} transferOp
   * @param {object} [params]
   * @param {string} [params.displayName]
   * @param {string} [params.module]
   * @param {string} [params.lex]
   * @param {Array} [params.subject] specified subject - otherwise will be
   * derived from the owner using `getSubject()`
   * @returns {Promise.<module:nmodule/webEditors/rc/wb/commands/transfer/TransferCommand>}
   * @see module:nmodule/webEditors/rc/wb/mixin/TransferSupport
   */
  TransferCommand.make = function (owner, transferOp, params) {
    var subject = params && params.subject;

    if (!owner || !mixinUtils.hasMixin(owner, 'transfer')) {
      return Promise.reject(new Error('owner with TransferSupport required'));
    }
    if (!transferOp) {
      return Promise.reject(new Error('transfer op required'));
    }

    return Promise.resolve(subject || owner.getSubject())
      .then(function (subject) {
        return new TransferCommand(owner, transferOp, _.extend({}, params, {
          subject: subject
        }));
      });
  };

  /**
   * Two `TransferCommands` will merge together if they are performing the same
   * transfer op on the subject.
   * @param {module:bajaux/commands/Command} cmd
   * @returns {module:nmodule/webEditors/rc/wb/commands/transfer/TransferCommand|null}
   */
  TransferCommand.prototype.merge = function (cmd) {
    if ((cmd instanceof TransferCommand) && (cmd.$op === this.$op)) {
      return this;
    }
    return null;
  };

  return TransferCommand;
});
