/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Tony Richards
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/links/LinkPad
 */
define(['baja!',
        'baja!baja:Component',
        'log!nmodule.webEditors.rc.wb.links.LinkPad',
        'bajaux/commands/Command',
        'bajaux/util/CommandButton',
        'lex!webEditors',
        'underscore',
        'Promise',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/wb/baja/linkCheckUtil',
        'nmodule/webEditors/rc/fe/feDialogs',
        'nmodule/webEditors/rc/fe/BaseWidget',
        'nmodule/webEditors/rc/fe/baja/util/typeUtils',
        'nmodule/webEditors/rc/wb/links/SlotChooser',
        'hbs!nmodule/webEditors/rc/wb/links/template/LinkPad'], function (
         baja,
         types,
         log,
         Command,
         CommandButton,
         lexs,
         _,
         Promise,
         fe,
         linkCheckUtil,
         feDialogs,
         BaseWidget,
         typeUtils,
         SlotChooser,
         tplLinkPad) {

  'use strict';

  var webEditorsLex = lexs[0],
      isComponent = typeUtils.isComponent,
      checkLink = linkCheckUtil.checkLink,
      logError = log.severe.bind(log);

  /**
   * Editor to link slots between components
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/BaseWidget
   * @alias module:nmodule/webEditors/rc/wb/links/LinkPad
   * @param {Object} params
   * @constructor
   */
  var LinkPad = function LinkPad(params) {
    var that = this;
    that.$isSwapped = false;

    BaseWidget.apply(that, arguments);

    that.$swapCommand = new Command({
      module: 'webEditors',
      lex: 'commands.linking.reverse',
      func: function() {
        return that.$handleSwapButton();
      }
    });
    that.getCommandGroup().add(that.$swapCommand);

    that.validators().add(function() {
      if(that.$getOutgoingSlotChooser().getSelectedSlot() &&
        that.$getIncomingSlotChooser().getSelectedSlot()) {
        return;
      }
      throw new Error(webEditorsLex.get('LinkPad.needSourceAndTarget'));
    });
  };
  LinkPad.prototype = Object.create(BaseWidget.prototype);
  LinkPad.prototype.constructor = LinkPad;

////////////////////////////////////////////////////////////////
// Private functions
////////////////////////////////////////////////////////////////

  /**
   * Get the swap command button
   *
   * @private
   * @returns {Command}
   */
  LinkPad.prototype.$getSwapCommand = function() {
    return this.$swapCommand;
  };

  /**
   * Get the element containing the specified slot chooser.
   *
   * @private
   * @param {Boolean} incoming
   * @returns {jQuery}
   */
  LinkPad.prototype.$getSlotChooserElement = function (incoming) {
    if (incoming) {
      return this.jq().find('.LinkPad-incoming-SlotChooser');
    } else {
      return this.jq().find('.LinkPad-outgoing-SlotChooser');
    }
  };

  /**
   * Get the element containing the incoming slot chooser.
   *
   * @private
   * @returns {jQuery}
   */
  LinkPad.prototype.$getIncomingElement = function () {
    return this.$getSlotChooserElement(!this.$isSwapped);
  };

  /**
   * Get the incoming `SlotChooser`.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/wb/links/SlotChooser}
   */
  LinkPad.prototype.$getIncomingSlotChooser = function () {
    return this.$getIncomingElement().data('widget');
  };

  /**
   * Get the element containing the outgoing slot chooser.
   *
   * @private
   * @returns {jQuery}
   */
  LinkPad.prototype.$getOutgoingElement = function () {
    return this.$getSlotChooserElement(this.$isSwapped);
  };

  /**
   * Get the outgoing `SlotChooser`.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/wb/links/SlotChooser}
   */
  LinkPad.prototype.$getOutgoingSlotChooser = function () {
    return this.$getOutgoingElement().data('widget');
  };

  /**
   * Create a filter and set it on specified SlotChooser
   *
   * The new filter will indicate if a link is valid based on the opposing `SlotChooser`
   * current selection.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/links/SlotChooser} filterTarget
   */
  LinkPad.prototype.$setFilter = function(filterTarget) {
    var that = this;

    return filterTarget.setEnableFilter(function (slot, batch) {
      var sourceSlot, targetSlot;

      if(filterTarget.$isTarget) {
        sourceSlot = that.$getOutgoingSlotChooser().$selected;
        targetSlot = slot;
      } else {
        sourceSlot = slot;
        targetSlot = that.$getIncomingSlotChooser().$selected;
      }

      if (!sourceSlot || !targetSlot) {
        return [true, null, null];
      }

      return checkLink({ source: that.$sources[0], target: that.$targets[0],
        sourceSlot: sourceSlot, targetSlot: targetSlot, batch: batch }, true).then( function(linkCheck) {
        if(linkCheck) {
          return [linkCheck.isValid(), linkCheck.getInvalidReason(), linkCheck.serverLinkCheck];
        }
        return false;
      });
    });
  };

  LinkPad.prototype.getSubscriber = function() {
    var that = this;
    if (!that.$subscriber) {
      that.$subscriber = new baja.Subscriber();
    }
    return that.$subscriber;
  };

////////////////////////////////////////////////////////////////
// LinkPad event handlers
////////////////////////////////////////////////////////////////

  /**
   * Event handler that is called when one of the child SlotChooser has a selection change.
   *
   * @private
   * @param {Slot} selected
   * @param {Slot} previouslySelected
   * @param {SlotChooser} slotChooser
   * @returns {Promise}
   */
  LinkPad.prototype.$handleChoice = function(selected, previouslySelected, slotChooser) {
    var that = this;

    if(slotChooser.$isTarget) {
      // If the slotChooser is a target, set the filter on the outgoing first,
      // and then set the filter on the incoming.
      return that.$setFilter(that.$getOutgoingSlotChooser()).then(function () {
        return that.$setFilter(that.$getIncomingSlotChooser());
      });
    } else {
      // If the slotChooser is not a target, set the filter on the incoming first,
      // and then set the filter on the outgoing.
      return that.$setFilter(that.$getIncomingSlotChooser()).then(function () {
        return that.$setFilter(that.$getOutgoingSlotChooser());
      });
    }
  };

  /**
   * Event handler that is called the swap button is pressed
   *
   * @private
  */
  LinkPad.prototype.$handleSwapButton = function() {
    var that = this,
        outChooser = that.$getOutgoingSlotChooser(),
        inChooser = that.$getIncomingSlotChooser();

    // Invert the isSwapped indicator
    that.$isSwapped = !that.$isSwapped;

    // Swap the columns on-screen using a flex box
    that.jq().find('.LinkPad-flex-container')
      .toggleClass('LinkPad-flex-reversed', that.$isSwapped);

    // Swap the targets
    inChooser.$isTarget = false;
    outChooser.$isTarget = true;

    // Swap the incoming / outgoing "Target" vs "Source" titles
    var outText, inText;
    if(that.$isSwapped) {
      inText = webEditorsLex.get('SlotChooser.source');
      outText = webEditorsLex.get('SlotChooser.target');
    } else {
      inText = webEditorsLex.get('SlotChooser.target');
      outText = webEditorsLex.get('SlotChooser.source');
    }

    that.jq().find('.LinkPad-outgoing-title-ts')
      .text(outText);
    that.jq().find('.LinkPad-incoming-title-ts')
      .text(inText);

    // Swap sources with targets;
    var tmp = that.$sources;
    that.$sources = that.$targets;
    that.$targets = tmp;

    // Re-filter in case a slot combination is no longer valid, preferring to
    // deselect the incoming slot chooser over the outgoing chooser.
    return that.$setFilter(that.$getIncomingSlotChooser()).then(function () {
      return that.$setFilter(that.$getOutgoingSlotChooser());
    });
  };

////////////////////////////////////////////////////////////////
// LinkPad bajaux implementation
////////////////////////////////////////////////////////////////

  /**
   * Initializes the incoming and outgoing slot choosers
   * @param {JQuery} dom
   * @returns {Promise} A promise to be resolved once the widget has
   * initialized
   */
  LinkPad.prototype.doInitialize = function (dom) {
    dom.html(tplLinkPad({
      source: webEditorsLex.get('SlotChooser.source'),
      target: webEditorsLex.get('SlotChooser.target')
    }))
      .addClass('LinkPad');

    var that = this,
        formFactor = that.getFormFactor(),
        props = that.properties(),
        readonly = that.isReadonly(),
        inReadonly = props.getValue('editIncoming') === false || readonly,
        outReadonly = props.getValue('editOutgoing') === false || readonly,
        outChooser = new SlotChooser({
          formFactor: formFactor, readonly: outReadonly
        }),
        inChooser = new SlotChooser({
          formFactor: formFactor, readonly: inReadonly
        });

    function eventHandler(selection, previousSelection, slotChooser) {
      that.setModified(true);
      that.$handleChoice(selection, previousSelection, slotChooser)
        .catch(logError);
    }

    outChooser.on('selectionChanged', eventHandler);
    inChooser.on('selectionChanged', eventHandler);
    inChooser.$isTarget = true;
    return Promise.all([
      fe.buildFor({
        value: that.$getSwapCommand(),
        dom: dom.find('.js-dialog-button-swap'),
        type: CommandButton
      }),
      outChooser.initialize(that.$getOutgoingElement()),
      inChooser.initialize(that.$getIncomingElement())
    ]);
  };

  /**
   * Implement the BaseWidget doLoad
   *
   * @param {Object} value
   * @param {Array.<baja.Component>} value.sources components that are being linked.
   * @param {Array.<baja.Component>} value.targets components that are being linked.
   *
   * @returns {Promise} resolves after sources and targets are loaded into their
   * corresponding slot choosers.
   */
  LinkPad.prototype.doLoad = function(value) {
    var that = this;

    if (!_.isObject(value)) { die('object required'); }

    var sources = value.sources, targets = value.targets;

    if (!_.isArray(sources) || !_.isArray(targets)) {
      die('LinkPad requires sources and targets specified');
    }

    var source = sources[0], target = targets[0], dom = that.jq();

    // TODO support multiple sources / targets
    if (!isComponent(source)) { die('LinkPad source must be a component'); }
    if (!isComponent(target)) { die('LinkPad target must be a component'); }

    that.$sources = sources;
    that.$targets = targets;

    dom.find('.LinkPad-outgoing-title-comp').text(source.getDisplayName());
    dom.find('.LinkPad-incoming-title-comp').text(target.getDisplayName());

    var subscriptions = that.getSubscriber().subscribe(sources.concat(targets));

    // Wait for the subscriptions to finish, then load sources and targets
    return Promise.all(subscriptions).then(function() {
      return Promise.all([
        that.$getOutgoingSlotChooser().load(sources),
        that.$getIncomingSlotChooser().load(targets)
      ]);
    });
  };

  /**
   * Implement the BaseWidget doRead function.
   *
   * @return {Object|null}
   */
  LinkPad.prototype.doRead = function () {
    var that = this;
    return Promise.join(
      that.$getIncomingSlotChooser().read(),
      that.$getOutgoingSlotChooser().read()
    ).spread(function (inOrd, outOrd) {
      if (inOrd && outOrd) {
        inOrd = inOrd.toString();
        outOrd = outOrd.toString();
        var ini = inOrd.lastIndexOf('/'),
            outi = outOrd.lastIndexOf('/'),
            sourceSlotName = outOrd.substr(outi + 1),
            targetSlotName = inOrd.substr(ini + 1);
        return {
          sources: that.$sources,
          targets: that.$targets,
          sourceOrd: that.$sources[0].getOrdInSession(),
          sourceSlot: sourceSlotName,
          targetSlot: targetSlotName
        };
      }
      // Return null if the in or out slot chooser returned null
      return null;
    });
  };

  /**
   * Remove LinkPad class and destroy slot choosers.
   * @returns {Promise}
   */
  LinkPad.prototype.doDestroy = function () {
    var that = this;
    that.jq().removeClass('LinkPad');

    that.getSubscriber().detach();

    return that.getChildWidgets().destroyAll()
      .then(function () {
        that.getSubscriber().unsubscribeAll().catch(logError);
        return null;
      });
  };

  /**
   * Show the LinkPad dialog, and if the OK button is clicked, create the lin and put
   * it on target `link` slot.
   *
   * @param {Object} params
   * @param {Array.<Component>} params.sources array of `baja:Component` objects that are
   * the sources objects to be linked.
   * @param {Array.<Component>} params.targets array of `baja:Component` objects that are
   * the target objects to be linked.
   * @param {object} [params.properties]
   * @param {boolean} [params.readonly]
   * @returns {Promise.<Link>} resolves to the newly created Link (or null)
   */
  LinkPad.doDialog = function(params) {
    return feDialogs.showFor({
      value: {
        sources: params.sources,
        targets: params.targets
      },
      properties: params.properties,
      readonly: params.readonly,
      type: LinkPad,
      formFactor: 'compact'
    }).then(function (result) {
      if(result) {
        // Check, make, add link (should this be here, or in the calling function?)
        var linkParams = {
          source: result.sources[0],
          sourceSlot: result.sourceSlot,
          targetSlot: result.targetSlot
        };
        // TODO Possibly should wrap all three of these calls into a single RPC
        return result.targets[0].checkLink(linkParams).then(function (linkCheck) {
          if(linkCheck.isValid()) {
            return result.targets[0].makeLink(linkParams)
              .then(function (link) {
                if(link) {
                  return result.targets[0].add({slot: 'Link?', value: link});
                }
              });
          }
        });
      }
    });
  };

  function die(msg) { throw new Error(msg); }

  return (LinkPad);
});
