/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Tony Richards
 */

/*jshint browser: true */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/links/SlotChooser
 */
define(['baja!',
        'jquery',
        'underscore',
        'Promise',
        'nmodule/webEditors/rc/fe/BaseWidget',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/fe/baja/IconEditor',
        'nmodule/webEditors/rc/fe/baja/util/slotUtils',
        'hbs!nmodule/webEditors/rc/wb/links/template/SlotChooser-row'], function (
         baja,
         $,
         _,
         Promise,
         BaseWidget,
         fe,
         IconEditor,
         slotUtils,
         tplSlotChooserRow) {

  'use strict';

  var toCssClass = slotUtils.toCssClass,
      getSlotIcon = slotUtils.getSlotIcon;

////////////////////////////////////////////////////////////////
// SlotChooser definition
////////////////////////////////////////////////////////////////

  /**
   * A panel that displays slots of a component and allows the user
   * to choose one (or more) of the slots.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/BaseWidget
   * @alias module:nmodule/webEditors/rc/wb/links/SlotChooser
   * @param {Object} [params] - See BaseWidget params
   */
  var SlotChooser = function SlotChooser(params) {
    // var that = this;
    BaseWidget.apply(this, arguments);

    this.$filter = function () { return true; };
    this.$selected = null;
  };

  SlotChooser.prototype = Object.create(BaseWidget.prototype);
  SlotChooser.constructor = SlotChooser;

  /**
   * @param {JQuery} dom
   */
  SlotChooser.prototype.doInitialize = function(dom) {
    var that = this;

    dom.addClass('SlotChooser');

    // Allow clicking from all SlotChooserRow, even if it's disabled; when clicking
    // on a disabled row, the opposing row becomes invalid, causing it to get
    // deselected.
    dom.on('click', '.SlotChooserRow', function (event) {
      // Don't continue bubbling the event up the DOM.
      event.preventDefault();

      that.setSelectedSlot($(this).data('slot'), true);
    });

    return BaseWidget.prototype.doInitialize.apply(that, arguments);
  };

  /**
   * Loads the slots for the specified components.  If more than one component is specified,
   *  the intersecting slots are loaded.
   *
   * @param {Array<baja.Component>} components an array of `baja:Component`
   * @returns {Promise|null} promise to be resolved with the requested components
   *  have been loaded
   */
  SlotChooser.prototype.doLoad = function(components) {
    var that = this;

    // These need to be leased at some point... but here / now?
    return baja.Component.lease(components).then(function () {
      that.$components = components;
      that.$slots = that.$getIntersectingSlots();
      return that.$loadSlots();
    });
  };

  SlotChooser.prototype.doDestroy = function() {
    this.jq().removeClass('SlotChooser');
  };

  /**
   * Sets the enable filter function to enable / disable slot selection.
   *
   * @param {function} filter function that is called once for each slot
   *  while (re)loading the slots into the DOM.  This filter function should
   *  return [true, ?, Promise.<LinkCheck>] to enable the slot for choosing, or
   *  [false, reason, Promise.<LinkCheck>] to disable the slot.
   *  reason is a string that is the reason the slot is not enabled
   *  Promise.<LinkCheck> (optional) is a Promise that resolve to a secondary LinkCheck.
   *  The first set of values in this return is "best guess" client-side link checking,
   *  whereas the LinkCheck third argument resolves to a definitive answer from the server.
   *  If the third element is null or missing then the first element Boolean is a definitive
   *  answer.
   */
  SlotChooser.prototype.setEnableFilter = function(filter) {
    this.$filter = filter;
    return this.$loadSlots();
  };

  /**
   * Reads the currently selected slot and turns it into an Ord.
   *
   * @returns {baja.Ord|null} the currently selected slot as a `baja:Ord`, or
   * null if none selected.
   */
  SlotChooser.prototype.doRead = function () {
    var selectedSlot = this.getSelectedSlot();
    if (selectedSlot) {
      return baja.Ord.make(selectedSlot);
    }
    return null;
  };

  /**
   * Set which slot is currently selected and optionally fire the setSelected event.
   *
   * @param {Slot} selected - newly selected slot
   * @param {boolean} [fireEvents=false] Optional true if the selectionChanged event
   * should be fired
   */
  SlotChooser.prototype.setSelectedSlot = function(selected, fireEvents) {
    var wasSelected = this.$selected;
    this.$selected = selected;
    if (fireEvents) {
      // Fire an event so the container can respond to the change.
      this.emit('selectionChanged', selected, wasSelected, this);

      // Mark the widget modified
      this.setModified(true);
    }
  };

  SlotChooser.prototype.$loadSlots = function () {
    var that = this,
      batch = new baja.comm.Batch(),
      rowPromises = that.$slots.filter(that.$isValidSlotRow.bind(that))
        .map(function(slot) {
          return that.$buildRowDom(slot, batch);
        });
    batch.commit();

    var unzipped = _.unzip(rowPromises),
        clientPromises = unzipped[0],
        serverPromises = unzipped[1];

    return Promise.all(clientPromises).then(function (rows) {
      that.jq().html(rows);
    }).then(function () {
      return Promise.all(serverPromises).then(function (rows) {
        that.jq().html(rows);
      });
    });
  };

  /**
   * Get the intersecting slots from all of the components.
   *
   * @private
   * @returns {Array} array of baja:Slot
   */
  SlotChooser.prototype.$getIntersectingSlots = function() {
    // TODO Implement; current implementation returns the slots from
    // the first component
    return this.$components[0].getSlots().toArray();
  };

  /**
   * @returns {baja.Slot|null} The currently selected slot, or null if no slot
   * has been selected
   */
  SlotChooser.prototype.getSelectedSlot = function () {
    return this.$selected;
  };

  /**
   * Creates a table row to hold an instance of `SlotChooserRow`.
   *
   * @private
   * @param {Slot} slot - Slot for which to build a row DOM.
   * @param {Batch} batch object for batching the server side link checks.
   *
   * @returns {Array.<Promise.<JQuery>>} the table row into which to load the field editor.
   * The first entry in the array resolves to the row dom based on the client link check,
   * and the second entry is based on the server (definitive) link check.
   */
  SlotChooser.prototype.$buildRowDom = function(slot, batch) {
    var that = this,
        buildRow = function(slot, enabled, reason, server) {
          var selected;
          if (slot === that.getSelectedSlot()) {
            // If the slot was previously selected but now it's no longer enabled
            // then mark it so that it's not selected. (but only do this if "enabled"
            // is coming from a server checkLink.
            if(!enabled && server) {
              that.setSelectedSlot(null, true);
            } else {
              selected = true;
            }
          }

          var name = that.$components[0].getDisplayName(slot),
            row = $(tplSlotChooserRow({
              slotClass: toCssClass(slot),
              slotType: getSlotType(slot),
              name: that.$components[0].getDisplayName(slot),
              enabled: enabled,
              selected: selected,
              tooltip: reason || name
            }));

          // Attach the slot to the row data to make it handy for referencing later
          row.data('slot', slot);

          return fe.buildFor({
            dom: row.children('.SlotChooserRow-icon'),
            value: getSlotIcon(slot),
            type: IconEditor
          })
            .then(function () {
              return row;
            });
        },
        clientPromise,
        serverPromise = new Promise(function (resolve, reject) {
          // Construct the HTML, specifying the slot style, name, icon
          clientPromise = Promise.resolve(that.$filter(slot, batch)).then(function (filterResponse) {
            var enabled, reason;

            if (Array.isArray(filterResponse)) {
              if(filterResponse.length > 1) {
                reason = filterResponse[1];
              } else {
                reason = '';
              }
              enabled = filterResponse[0];
              if(filterResponse.length > 2 && filterResponse[2]) {
                  resolve(filterResponse[2].then(function (checkLink) {
                    return buildRow(slot, checkLink.isValid(), checkLink.getInvalidReason(), true);
                  }));
              } else {
                resolve(buildRow(slot, enabled, reason));
              }
            } else {
              enabled = filterResponse;
              resolve(buildRow(slot, enabled, reason));
            }

            return buildRow(slot, enabled, reason);
          });
        });

    return [ clientPromise, serverPromise ];
  };

  /**
   * Checks to see if the specified slot is a valid slot row
   *
   * @private
   * @param {baja.Slot} slot to determine if it should be displayed in this chooser
   * @returns {boolean} true if the slot is valid, otherwise false
   */
  SlotChooser.prototype.$isValidSlotRow = function(slot) {
    var that = this;

    // Not a valid slot if it's hidden
    if(slot.getFlags() & baja.Flags.HIDDEN) {
      return false;
    }

    if (slot.isProperty()) {
      // Not valid if it's a Component (unless it's a Vector), Link, or WsAnnotation
      var v = that.$components[0].get(slot);

      if (baja.hasType(v, 'baja:Component') && !baja.hasType(v, 'baja:Vector')) {
        return false;
      }
      // TODO If a Link is not valid, why would a Relation be valid?
      if (baja.hasType(v, 'baja:Link')) {
        return false;
      }
      if (baja.hasType(v, 'baja:WsAnnotation')) {
        return false;
      }
      return true;
    } else {
      return true;
    }
  };

  /**
   * Get slot type for styling purposes - action, topic, relation, or property.
   * @param {baja.Slot} slot
   * @returns {string}
   */
  function getSlotType(slot) {
    if (slot.isAction()) { return 'action'; }
    if (slot.isTopic()) { return 'topic'; }
    if (slot.getType().is('baja:Relation')) { return 'relation'; }
    return 'property';
  }

  return (SlotChooser);
});

