/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/menu/componentMenuAgent
 */
define(['baja!',
        'bajaux/commands/CommandGroup',
        'Promise',
        'underscore',
        'nmodule/js/rc/asyncUtils/asyncUtils',
        'nmodule/webEditors/rc/wb/commands/transfer/TransferCommand',
        'nmodule/webEditors/rc/wb/menu/menuUtils',
        'nmodule/webEditors/rc/wb/mixin/mixinUtils'], function (
         baja,
         CommandGroup,
         Promise,
         _,
         asyncUtils,
         TransferCommand,
         menuUtils,
         mixinUtils) {

  'use strict';

  var doRequire = asyncUtils.doRequire,
      ifAllowedByProfile = menuUtils.ifAllowedByProfile,
      makeActionsGroup = menuUtils.makeActionsGroup,
      makeNewGroup = menuUtils.makeNewGroup,
      makeViewsGroup = menuUtils.makeViewsGroup,
      hasMixin = mixinUtils.hasMixin;

  function isParented(comp) {
    return !!comp.getPropertyInParent();
  }

  /**
   * Check to see if the component is frozen (either unparented, or a frozen
   * slot).
   *
   * @inner
   * @param {baja.Complex} comp
   * @returns {Boolean}
   */
  function isFrozen(comp) {
    var prop = comp.getPropertyInParent();
    return !prop || prop.isFrozen();
  }

  function isReadonly(comp) {
    var space = comp.getComponentSpace();
    return space && space.isSpaceReadonly();
  }

  /**
   * Asynchronously require the appropriate command modules and instantiate
   * the `CommandGroup`. Assign the appropriate enabled/disabled status for
   * the different commands depending on the write/invoke permissions on this
   * component.
   *
   * @inner
   * @param {*} owner
   * @param {baja.Component} comp
   * @returns {Promise}
   */
  function makeGroup(owner, comp) {
    return doRequire([
      'nmodule/webEditors/rc/wb/commands/AddSlotCommand',
      'nmodule/webEditors/rc/wb/commands/RenameCommand',
      'nmodule/webEditors/rc/wb/commands/LinkMarkCommand',
      'nmodule/webEditors/rc/wb/commands/LinkFromCommand',
      'nmodule/webEditors/rc/wb/commands/LinkToCommand',
      'nmodule/webEditors/rc/wb/commands/ReorderSlotsCommand',
      'nmodule/webEditors/rc/wb/commands/SetDisplayNameCommand',
      'nmodule/webEditors/rc/wb/commands/CompositeEditorCommand'
    ])
    .spread(function (
       AddSlotCommand,
       RenameCommand,
       LinkMarkCommand,
       LinkFromCommand,
       LinkToCommand,
       ReorderSlotsCommand,
       SetDisplayNameCommand,
       CompositeEditorCommand) {

      var group = new CommandGroup(),
          newGroup,
          deleteCmd, renameCmd, setDisplayNameCmd, reorderCmd, compositeCmd;

      if (!isReadonly(comp)) {
        var frozen = isFrozen(comp),
            parented = isParented(comp),
            enabled = comp.getPermissions().hasAdminWrite();

        newGroup = makeNewGroup(comp, new AddSlotCommand(comp));
        setDisplayNameCmd = new SetDisplayNameCommand(comp);
        reorderCmd = new ReorderSlotsCommand(comp);
        compositeCmd = new CompositeEditorCommand(comp);

        setDisplayNameCmd.setEnabled(enabled && parented);
        reorderCmd.setEnabled(enabled);
        compositeCmd.setEnabled(enabled && parented);

        if (hasMixin(owner, 'transfer')) {
          var getSubject = Promise.resolve(owner.getSubject());

          deleteCmd = getSubject
            .then(function (subject) {
              if (owner.isTransferOpSupported('delete', subject)) {
                return TransferCommand.make(owner, 'delete', {
                  subject: subject,
                  module: 'webEditors',
                  lex: 'commands.delete'
                })
                  .then(function (cmd) {
                    if (!enabled || frozen) { cmd.setEnabled(false); }
                    return cmd;
                  });
              }
            });

          renameCmd = getSubject
            .then(function (subject) {
              if (owner.isTransferOpSupported('rename', subject)) {
                return TransferCommand.make(owner, 'rename', {
                  subject: subject,
                  module: 'webEditors',
                  lex: 'commands.rename'
                })
                  .then(function (cmd) {
                    if (!enabled || frozen) { cmd.setEnabled(false); }
                    return cmd;
                  });
              }
            });
        } else {
          renameCmd = new RenameCommand(comp);
          renameCmd.setEnabled(enabled && !frozen);
        }
      }

      //TODO: TransferWidget.isDeleteEnabled()?
      return Promise.all([
        ifAllowedByProfile('viewSelection', function () {
          return makeViewsGroup(comp.getNavOrd());
        }),
        newGroup,
        makeActionsGroup(comp),
        deleteCmd,
        renameCmd,
        setDisplayNameCmd,
        reorderCmd,
        LinkMarkCommand.make(comp),
        LinkFromCommand.make(comp),
        LinkToCommand.make(comp),
        compositeCmd
      ])
        .then(function (allKids) {
          group.add.apply(group, _.compact(allKids));

          return group;
        });
    });
  }

  /**
   * Create a command group that can be used to construct a context menu. Note
   * that the component, if mounted, will be leased at this time so any
   * dynamic actions can be included in the menu.
   *
   * The value may also be a `NavContainer`, e.g., a nav node corresponding to
   * a component inside a .bog file inside the station's file space, that
   * BajaScript has not fully resolved yet. In this case, the `NavContainer`'s
   * nav ORD will be resolved and the menu will be generated from the result.
   *
   * @private
   * @exports nmodule/webEditors/rc/wb/menu/componentMenuAgent
   * @param {*} owner if the owner has the `TransferSupport` mixin, then
   * transfer commands will be included
   * @param {baja.Component|baja.NavContainer} comp
   * @returns {Promise} promise to be resolved with a `CommandGroup`
   * for the given component
   * @see module:nmodule/webEditors/rc/wb/mixin/TransferSupport
   */
  var componentMenuAgent = function componentMenuAgent(owner, comp) {
    if (comp instanceof baja.NavContainer) {
      return comp.getNavOrd().get({ lease: true })
        .then(function (comp) {
          return makeGroup(owner, comp);
        });
    }

    return Promise.resolve(comp.isMounted() && comp.lease())
      .then(function () {
        return makeGroup(owner, comp);
      });
  };

  return (componentMenuAgent);
});
