/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/mgr/model/MgrModelCompositeBuilder
 */
define(['baja!',
        'Promise',
        'nmodule/webEditors/rc/fe/config/CompositeBuilder',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/wb/mgr/model/MgrModel'], function (
         baja,
         Promise,
         CompositeBuilder,
         fe,
         MgrModel) {

  'use strict';

  function getColumnName(column) { return column.getName(); }

  /**
   * `CompositeBuilder` that watches selected rows in a `MgrModel`. It delegates
   * the construction of editors and the retrieval of values to the individual
   * `MgrColumn`s in the `MgrModel`.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/mgr/model/MgrModelCompositeBuilder
   * @extends module:nmodule/webEditors/rc/fe/config/CompositeBuilder
   * @param {Object} params
   * @param {module:nmodule/webEditors/rc/wb/mgr/model/MgrModel} params.model
   * the `MgrModel` to be used as this builder's data source
   * @param {module:nmodule/webEditors/rc/util/ListSelection} params.selection
   * the `ListSelection` that tracks which rows in the `MgrModel` are selected.
   */
  var MgrModelCompositeBuilder = function MgrModelCompositeBuilder(params) {
    var model = params && params.model,
        selection = params && params.selection;

    if (!selection) {
      throw new Error('ListSelection required');
    }

    this.$selection = selection;

    CompositeBuilder.call(this, { dataSource: model });
  };
  MgrModelCompositeBuilder.prototype = Object.create(CompositeBuilder.prototype);
  MgrModelCompositeBuilder.prototype.constructor = MgrModelCompositeBuilder;

  /**
   * Get the column matching the given name in the `MgrModel` data source.
   *
   * @private
   * @param {String} columnName
   * @returns {module:nmodule/webEditors/rc/wb/mgr/model/MgrColumn} the
   * `MgrColumn` with the given name, or `null` if not found
   */
  MgrModelCompositeBuilder.prototype.$getColumn = function (columnName) {
    return this.getDataSource().getColumn(columnName);
  };

  /**
   * Get all selected rows in the `MgrModel` data source as determined by the
   * builder's `ListSelection`.
   *
   * @private
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>}
   */
  MgrModelCompositeBuilder.prototype.$getSelectedRows = function () {
    var model = this.getDataSource(),
        selection = this.$selection;

    return selection.getSelectedElements(model.getRows());
  };

  /**
   * Set the `MgrModel` data source.
   *
   * @param {module:nmodule/webEditors/rc/wb/mgr/model/MgrModel} model
   * @returns {Promise} promise to be resolved with the `MgrModel` has
   * been set as the data source
   */
  MgrModelCompositeBuilder.prototype.setDataSource = function (model) {
    if (model && !(model instanceof MgrModel)) {
      return Promise.reject(new Error('MgrModel required'));
    }

    return CompositeBuilder.prototype.setDataSource.apply(this, arguments);
  };

  /**
   * Delegates to the matching `MgrColumn`'s `getConfigFor` function, using
   * the currently selected `Row`s.
   *
   * @param {String} key the column name
   * @returns {Object} a config object to be given to `fe.makeFor`
   */
  MgrModelCompositeBuilder.prototype.getConfigFor = function (key) {
    var column = this.$getColumn(key);

    return column && column.getConfigFor(this.$getSelectedRows());
  };

  /**
   * Resolve the display name from the matching column.
   *
   * @param {String} key the column name
   * @returns {Promise} promise to be resolved with the display name,
   * or with `null` if column not found
   */
  MgrModelCompositeBuilder.prototype.getDisplayNameFor = function (key) {
    var column = this.$getColumn(key);

    return Promise.resolve(column && column.toDisplayName());
  };

  /**
   * Resolve the column icon of the matching column.
   *
   * @param {String} key the column name
   * @returns {Promise} promise to be resolved with an icon URI, or `null`
   * if column not found
   */
  MgrModelCompositeBuilder.prototype.getIconFor = function (key) {
    var column = this.$getColumn(key);
    return Promise.resolve(column && column.getColumnIcon());
  };

  /**
   * Get the names of all editable columns in the `MgrModel`.
   *
   * @returns {Promise} promise to be resolved with an array of string names
   */
  MgrModelCompositeBuilder.prototype.getKeys = function () {
    var model = this.getDataSource();

    return Promise.resolve(model.getEditableColumns().map(getColumnName));
  };

  /**
   * Get a value coalesced from the matching column, using the currently
   * selected rows.
   *
   * @param {String} key the column name
   * @returns {*} a value given to us by the column's `coalesceRows` function,
   * using the currently selected rows - or `null` if column not found
   */
  MgrModelCompositeBuilder.prototype.getValueFor = function (key) {
    var column = this.$getColumn(key);

    return column && column.coalesceRows(this.$getSelectedRows());
  };


  /**
   * If the editor for this key has been built, and is still considered by the
   * column for this key to be suitable for the selected rows, just reuse
   * that editor.
   *
   * Otherwise, rebuild a new editor for this key.
   *
   * @private
   * @param {String} key
   * @returns {Promise} promise to be resolved with an editor the referenced
   * column likes for the selected rows.
   */
  MgrModelCompositeBuilder.prototype.$makeFor = function (key) {
    var that = this,
        ed = that.getEditorFor(key),
        column = that.$getColumn(key);

    if (ed &&
      ed.isInitialized() &&
      column.isEditorSuitable(getValueEditor(ed), this.$getSelectedRows())) {
      return Promise.resolve(ed);
    }

    return Promise.resolve(ed && ed.isInitialized() && ed.destroy())
      .then(function () {
        return that.getConfigFor(key);
      })
      .then(function (config) {
        //TODO: remove readonly checks after password rpc works offline
        if (baja.isOffline()) { config.readonly = true; }
        return fe.makeFor(config);
      })
      .then(function (ed) {
        return (that.$map[key] = ed);
      });
  };

  //ed can be a PropertySheetRow, which custom columns shouldn't have to know
  //about. pass the value editor instead.
  //TODO: this is probably not the best way to go about this
  function getValueEditor(ed) {
    if (ed.$getValueEditor) {
      return ed.$getValueEditor();
    }
    return ed;
  }

  return (MgrModelCompositeBuilder);
});
