/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @module nmodule/webEditors/rc/wb/table/model/TableModel
 */
define(['Promise',
        'underscore',
        'nmodule/js/rc/switchboard/switchboard',
        'nmodule/js/rc/tinyevents/tinyevents',
        'nmodule/webEditors/rc/wb/table/model/Column',
        'nmodule/webEditors/rc/wb/table/model/Row'], function (
         Promise,
         _,
         switchboard,
         tinyevents,
         Column,
         Row) {

  'use strict';

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  /** @param {String} err */
  function reject(err) { return Promise.reject(new Error(err)); }

  /**
   * @inner
   * @param {Array} arr array we want to remove from
   * @param {Array} toRemove array of things to remove
   * @returns {Promise} promise to be resolved if remove can be performed
   */
  function validateArrayToRemove(arr, toRemove) {
    for (var i = 0; i < toRemove.length; i++) {
      if (arr.indexOf(toRemove[i]) < 0) {
        return reject('cannot remove a row or column not already in model');
      }
    }
    return Promise.resolve();
  }

  /**
   * @inner
   * @param {Array} arr array we want to remove from
   * @param {Number} start start removing here (inclusive)
   * @param {Number} end end removing here (exclusive)
   * @returns {Promise} promise to be resolved if remove can be performed
   */
  function validateIndicesToRemove(arr, start, end) {
    if (start < 0 ||
      start >= arr.length ||
      'number' !== typeof end ||
      end < start ||
      end > arr.length) {

      return reject('invalid range to remove ' + start + ' - ' + end);
    }
    return Promise.resolve();
  }

  /**
   * @inner
   * @param {Array} arr array we want to remove from
   * @param {Array} toRemove array of things to remove
   * @returns {Promise} promise to be resolved with an array if remove
   * is successful, index 0 is the new array after removing, index 1 is the
   * array of objects removed
   */
  function removeArrayMembers(arr, toRemove) {
    return validateArrayToRemove(arr, toRemove)
      .then(function () {
        var removed = [],
            indices = [],
            newArr = arr.filter(function (thing, i) {
              if (toRemove.indexOf(thing) >= 0) {
                removed.push(thing);
                indices.push(i);
                return false;
              }
              return true;
            });

        return [ newArr, removed, indices ];
      });
  }

  /**
   * @inner
   * @param {Array} arr array to remove from
   * @param {Number} start start removing here (inclusive)
   * @param {Number} end end removing here (exclusive)
   * @returns {Promise} promise to be resolved with an array if remove
   * is successful, index 0 is the new array after removing, index 1 is the
   * array of objects removed
   */
  function removeArrayIndices(arr, start, end) {
    return validateIndicesToRemove(arr, start, end)
      .then(function () {
        var removed = arr.splice(start, end - start);
        return [ arr, removed, _.range(start, end) ];
      });
  }

  /**
   * @inner
   * @param {Array} arr array to remove from
   * @param {Array|Number} toRemove array of things to remove; or, start index
   * @param {Number} [end] end index
   * @returns {Promise} promise to be resolved with an array if remove
   * is successful, index 0 is the new array after removing, index 1 is the
   * array of objects removed
   */
  function doArrayRemove(arr, toRemove, end) {
    if (Array.isArray(toRemove)) {
      return removeArrayMembers(arr, toRemove);
    } else if (typeof toRemove === 'number') {
      return removeArrayIndices(arr, toRemove, end);
    } else {
      return reject('could not determine members to remove');
    }
  }

  /**
   * @inner
   * @param {Array} arr array to insert into
   * @param {Array} toInsert array of things to insert
   * @param {Number} [index] index to insert at (will just append to the end if
   * omitted)
   * @returns {Promise} promise to be resolved with the new array if
   * insert is successful
   */
  function doArrayInsert(arr, toInsert, index) {
    var len = arr.length;

    if (typeof index === 'number') {
      if (index < 0 || index > len) {
        return reject('index out of range: ' + index);
      }
    } else {
      index = len;
    }

    Array.prototype.splice.apply(arr, [ index, 0 ].concat(toInsert));
    return Promise.resolve([ arr, index ]);
  }

  /**
   * @inner
   * @param {module:nmodule/webEditors/rc/wb/table/model/TableModel} tableModel
   * model we are inserting columns into
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>} columns
   * @returns {Promise} promise to be resolved if columns can be inserted
   */
  function validateColumnsToInsert(tableModel, columns) {
    if (!Array.isArray(columns)) {
      return reject('array required');
    }

    for (var i = 0; i < columns.length; i++) {
      if (!(columns[i] instanceof Column)) {
        return reject('only Columns can be added');
      }
    }

    return Promise.resolve(columns);
  }

  /**
   * @inner
   * @param {module:nmodule/webEditors/rc/wb/table/model/TableModel} tableModel
   * model we are inserting rows into
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Row|*>} rows
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>} array
   * of rows
   * @throws {Error} if no rows given
   */
  function validateRowsToInsert(tableModel, rows) {
    if (!Array.isArray(rows)) {
      throw new Error('array of rows/subjects required');
    }

    return _.map(rows, function (row) {
      return row instanceof Row ? row : tableModel.makeRow(row);
    });
  }

////////////////////////////////////////////////////////////////
// TableModel
////////////////////////////////////////////////////////////////

  /**
   * API Status: **Development**
   *
   * Table Model, for use in backing a `Table` widget or similar.
   *
   * @class
   * @alias module:nmodule/webEditors/rc/wb/table/model/TableModel
   * @mixes tinyevents
   * @param {Object} [params]
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>} [params.columns]
   * @param {Array} [params.rows] if given, the values will be converted to
   * the model's initial rows by passing them through `makeRow()`.
   */
  var TableModel = function TableModel(params) {
    params = params || {};

    var that = this,
        columns = (params.columns || []).slice(),
        rows = params.rows || [],
        flagsChangedHandler = function (flags) {
          that.emit('columnsFlagsChanged', [this], [flags]);
        };

    tinyevents(that);

    _.each(columns, function (column) {
      column.on('flagsChanged', flagsChangedHandler);
    });

    that.$flagsChangedHandler = flagsChangedHandler;

    that.$columns = columns;
    that.$rows = validateRowsToInsert(that, rows);

    //ensure these async methods don't step on each other if called in quick succession.
    var queueUp = { allow: 'oneAtATime', onRepeat: 'queue' };
    switchboard(that, {
      insertRows: queueUp,
      insertColumns: queueUp,
      removeRows: queueUp,
      removeColumns: queueUp
    });
  };

  /**
   * Add new columns to the model. Will trigger a `columnsAdded` `tinyevent`.
   *
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>} toInsert
   * @param {Number} [index] index to insert the columns; will append to the
   * end if omitted
   * @returns {Promise} promise to be resolved if the insert is
   * successful
   */
  TableModel.prototype.insertColumns = function (toInsert, index) {
    var that = this,
        flagsChangedHandler = that.$flagsChangedHandler;

    return validateColumnsToInsert(that, toInsert)
      .then(function (columnsToInsert) {

        _.each(columnsToInsert, function (column) {
          column.on('flagsChanged', flagsChangedHandler);
        });

        return doArrayInsert(that.$columns, columnsToInsert, index)
          .spread(function (newColumns, index) {
            that.$columns = newColumns;
            that.emit('columnsAdded', columnsToInsert, index);
            return null; // squelch "promise not returned" warnings
          });
      });
  };

  /**
   * Remove columns from the model. Will trigger a `columnsRemoved` `tinyevent`.
   *
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>|Number} toRemove
   * the columns to remove; or, start index
   * @param {Number} [end] end index
   * @returns {Promise} promise to be resolved if the remove is
   * successful
   */
  TableModel.prototype.removeColumns = function (toRemove, end) {
    var that = this,
        flagsChangedHandler = that.$flagsChangedHandler;

    return doArrayRemove(that.getColumns(), toRemove, end)
      .then(function (arr) {
        var newColumns = arr[0],
            removed = arr[1];

        _.each(removed, function (column) {
          column.removeListener('flagsChanged', flagsChangedHandler);
        });

        that.$columns = newColumns;
        that.emit('columnsRemoved', removed);
        return null; // squelch "promise not returned" warnings
      });
  };

  /**
   * Get the column in this model matching the given name.
   *
   * @param {String} name
   * @returns {module:nmodule/webEditors/rc/wb/table/model/Column} the matching
   * column, or `null` if not found
   */
  TableModel.prototype.getColumn = function (name) {
    if (!name) {
      return null;
    }

    return _.find(this.$columns, function (column) {
      return column.getName() === name;
    }) || null;
  };

  /**
   * Get the current set of columns, optionally filtered by flags.
   *
   * @param {Number} [flags] if given, only return columns that have these
   * flags.
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>}
   */
  TableModel.prototype.getColumns = function (flags) {
    return this.$columns.filter(function (col) {
      return !flags || col.hasFlags(flags);
    });
  };

  /**
   * Get the index of the given column.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} column
   * @returns {number} the column's index, or -1 if not found
   */
  TableModel.prototype.getColumnIndex = function (column) {
    return _.indexOf(this.$columns, column);
  };

  /**
   * Return all columns with the `EDITABLE` flag set.
   *
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Column>}
   */
  TableModel.prototype.getEditableColumns = function () {
    return this.getColumns(Column.flags.EDITABLE);
  };

  /**
   * Ask the column at the given index for the value from the row at the
   * given index.
   *
   * @param {Number} x column index
   * @param {Number} y row index
   * @returns {Promise} promise to be resolved with the value
   */
  TableModel.prototype.getValueAt = function (x, y) {
    var row = this.$rows[y],
        column = this.$columns[x];

    return !row ? reject('row not found at ' + y) :
      !column ? reject('column not found at ' + x) :
      Promise.resolve(column.getValueFor(row));
  };

  /**
   * Instantiate a new row for the given subject. `insertRows` will delegate
   * to this if values are passed in rather than `Row` instances. Override
   * as necessary.
   *
   * @param {*} subject
   * @returns {module:nmodule/webEditors/rc/wb/table/model/Row}
   */
  TableModel.prototype.makeRow = function (subject) {
    return new Row(subject);
  };

  /**
   * Add new rows to the model. If non-`Row` instances are given, they will be
   * converted to `Row`s using `makeRow()`.
   *
   * Will trigger a `rowsAdded` `tinyevent`.
   *
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Row|*>} toInsert
   * @param {Number} [index] index to insert the rows; will append to the
   * end if omitted
   * @returns {Promise} promise to be resolved if the insert is
   * successful
   */
  TableModel.prototype.insertRows = function (toInsert, index) {
    var that = this,
        rowsToInsert;

    try {
      rowsToInsert = validateRowsToInsert(that, toInsert);
    } catch (e) {
      return reject(e);
    }

    return doArrayInsert(that.$rows, rowsToInsert, index)
      .spread(function (newRows, index) {
        that.$rows = newRows;
        that.emit('rowsAdded', rowsToInsert, index);
        return null; // squelch "promise not returned" warnings
      });
  };

  /**
   * Remove rows from the model. Will trigger a `rowsRemoved` `tinyevent`, with
   * parameters:
   *
   * - `rowsRemoved`: the rows that were removed
   * - `indices`: the original indices of the rows that were removed
   *
   * Note that `rowsRemoved` and `indices` will always be sorted by their
   * original index in the model's rows, regardless of the order of rows passed
   * to the `removeRows` function.
   *
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>|Number} toRemove
   * the rows to remove; or, start index
   * @param [end] end index
   * @returns {Promise} promise to be resolved if the remove is
   * successful
   */
  TableModel.prototype.removeRows = function (toRemove, end) {
    var that = this;

    return doArrayRemove(that.getRows(), toRemove, end)
      .spread(function (newRows, rowsRemoved, indices) {
        that.$rows = newRows;
        that.emit('rowsRemoved', rowsRemoved, indices);
        return null; // squelch "promise not returned" warnings
      });
  };

  /**
   * Get the current set of rows.
   *
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>}
   */
  TableModel.prototype.getRows = function () {
    return this.$rows.slice();
  };

  /**
   * Get the index of the given row.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/model/Row} row
   * @returns {number} the row's index, or -1 if not found
   */
  TableModel.prototype.getRowIndex = function (row) {
    return _.indexOf(this.$rows, row);
  };

  /**
   * Sort the table's rows according to the given sort function. Emits a
   * `rowsReordered` event.
   *
   * Remember that `Array#sort` is synchronous, so if the sort needs to use
   * any data that is asynchronously retrieved, the async work must be performed
   * *before* the sort so that the sort function can work synchronously.
   *
   * @param {Function} sortFunction standard array sort function to receive
   * two `Row` instances
   * @throws {Error} if a non-Function is given
   */
  TableModel.prototype.sort = function (sortFunction) {
    if (typeof sortFunction !== 'function') {
      throw new Error('sort function required');
    }

    this.$rows.sort(sortFunction);
    this.emit('rowsReordered');
  };

  return TableModel;
});
