/*
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/table/tree/TreeTable
 */
define(['bajaux/commands/Command',
        'bajaux/util/CommandButton',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/wb/table/Table',
        'nmodule/webEditors/rc/wb/table/tree/TreeTableModel'], function (
         Command,
         CommandButton,
         $,
         Promise,
         _,
         fe,
         Table,
         TreeTableModel) {

  'use strict';

  var SPACER_HTML = '<div class="spacer"/>';

  function makeSpacers(count) {
    return _.map(_.range(0, count), _.constant(SPACER_HTML)).join('');
  }

  /**
   * A table backed by a
   * {@link module:nmodule/webEditors/rc/wb/table/tree/TreeTableModel TreeTableModel},
   * that provides buttons to expand/collapse sections of the tree.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/table/tree/TreeTable
   * @extends module:nmodule/webEditors/rc/wb/table/Table
   */
  var TreeTable = function TreeTable() {
    Table.apply(this, arguments);
  };
  TreeTable.prototype = Object.create(Table.prototype);
  TreeTable.prototype.constructor = TreeTable;

  /**
   * Adds `TreeTable` class.
   *
   * @param {JQuery} dom
   * @returns {Promise}
   */
  TreeTable.prototype.doInitialize = function (dom) {
    dom.addClass('TreeTable');
    return Table.prototype.doInitialize.apply(this, arguments);
  };

  /**
   * Loads in a `TreeTableModel`.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/tree/TreeTableModel} tableModel
   * @returns {Promise}
   */
  TreeTable.prototype.doLoad = function (tableModel) {
    if (!(tableModel instanceof TreeTableModel)) {
      throw new Error('TreeTableModel required');
    }
    return Table.prototype.doLoad.apply(this, arguments);
  };

  /**
   * Remove the `TreeTable` class and destroy child widgets.
   *
   * @returns {Promise}
   */
  TreeTable.prototype.doDestroy = function () {
    var that = this;
    that.jq().removeClass('TreeTable');
    return that.getChildWidgets().destroyAll()
      .then(function () {
        return Table.prototype.doDestroy.apply(that, arguments);
      });
  };

  /**
   * Build the cell for the specified column and row. For the first column,
   * indent to show the tree depth, and add a button to expand/collapse the
   * tree at that row.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} column
   * @param {module:nmodule/webEditors/rc/wb/table/model/Row} row
   * @param {JQuery} dom
   * @returns {Promise}
   */
  TreeTable.prototype.buildCell = function (column, row, dom) {
    var that = this,
        model = that.getModel();
    return Promise.resolve(Table.prototype.buildCell.apply(that, arguments))
      .then(function () {
        if (model.getColumnIndex(column) !== 0) { return; } //first column only

        var prom;
        if (model.isExpandable(row)) {
          var buttonDom = $('<button class="expand" type="button"/>')
            .addClass(model.isExpanded(row) ? 'expanded' : 'collapsed')
            .prependTo(dom);

          prom = fe.buildFor({
            type: CommandButton,
            dom: buttonDom,
            value: new Command({
              func: function () {
                var expanded = model.isExpanded(row);
                buttonDom
                  .toggleClass('expanded', !expanded)
                  .toggleClass('collapsed', expanded);
                return model.toggle(row);
              }
            })
          });
        }

        dom.prepend(makeSpacers(model.getDepth(row)));

        return prom;
      });
  };

  return TreeTable;
});
