/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/util/RolesEditor
 */
define(['baja!',
        'baja!baja:Role',
        'bajaux/Widget',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/CompositeEditor',
        'nmodule/webEditors/rc/fe/baja/util/spaceUtils',
        'nmodule/webEditors/rc/fe/baja/util/subscriberUtils',
        'nmodule/webEditors/rc/fe/config/CompositeBuilder'], function (
         baja,
         types,
         Widget,
         $,
         Promise,
         _,
         CompositeEditor,
         spaceUtils,
         subscriberUtils,
         CompositeBuilder) {

  'use strict';

  var MAX_ROLES = 10,
      resolveService = spaceUtils.resolveService,
      leaseWhenNeeded = subscriberUtils.leaseWhenNeeded,
      getRoleService = leaseWhenNeeded(_.partial(resolveService, 'baja:RoleService')),
      getUserService = leaseWhenNeeded(_.partial(resolveService, 'baja:UserService'));

  /**
   * Editor for editing available roles on a user.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/util/RolesEditor
   * @extends module:nmodule/webEditors/rc/fe/CompositeEditor
   */
  var RolesEditor = function RolesEditor(params) {
    CompositeEditor.apply(this, arguments);
  };
  RolesEditor.prototype = Object.create(CompositeEditor.prototype);
  RolesEditor.prototype.constructor = RolesEditor;

  RolesEditor.prototype.$getRoleService = function () {
    return this.getOrdBase().then(getRoleService);
  };

  RolesEditor.prototype.$getUserService = function () {
    return this.getOrdBase().then(getUserService);
  };

  RolesEditor.prototype.$getCurrentUser = function () {
    return this.$getUserService()
      .then(function (userService) {
        var user = userService.get(baja.getUserName());
        return user && user.lease().then(_.constant(user));
      })
      .catch(_.constant(null));
  };

  /**
   * Resolve the roles that are available to add to a user.
   *
   * @private
   * @returns {Promise.<Array.<String>>} promise to be resolved with an array
   * of role names
   */
  RolesEditor.prototype.$getAvailableRoles = function () {
    var that = this;

    return Promise.join(that.$getRoleService(), that.$getCurrentUser())
      .spread(function (roleService, user) {
        var roleNames = _.map(roleService.getSlots().is('baja:IRole').toArray(),
          String);

        // remove all roles that don't belong to the logged in user.

        // super user can set any roles. If we couldn't find the user, just
        // show all, the station side check will prevent setting roles that
        // they aren't allowed.
        if (!user || user.get('permissions').isSuperUser()) {
          return roleNames;
        }

        // only add roles that were already assigned to the target user
        // or are assigned to the logged in user
        var allowedRoles = user.get('roles').split(','),
            loadedRoles = that.value().split(',');
        return _.filter(roleNames, function (role) {
          return _.contains(allowedRoles, role) || _.contains(loadedRoles, role);
        });
      })
      .catch(function () {
        return [];
      });
  };

  RolesEditor.prototype.$getRolesContainerElement = function () {
    return this.jq().children('.rolesContainer');
  };


  /**
   * Add "RolesEditor" class.
   *
   * @param {JQuery} dom
   */
  RolesEditor.prototype.doInitialize = function (dom) {
    dom.addClass('RolesEditor');
    dom.append('<div class="rolesContainer"/>');
  };

  /**
   * The builder for a `RolesEditor` will create a boolean editor for each
   * available role.
   *
   * @returns {module:nmodule/webEditors/rc/fe/config/CompositeBuilder}
   */
  RolesEditor.prototype.makeBuilder = function () {
    var that = this,
        builder = new CompositeBuilder();

    builder.getKeys = function () {
      return that.$getAvailableRoles()
        .then(function (roleNames) {
          var loadedRoles = that.value();

          if (that.getFormFactor() === Widget.formfactor.mini &&
            typeof loadedRoles === 'string') {

            loadedRoles = loadedRoles.split(',');

            for (var i = roleNames.length - 1; i >= 0; i--) {
              if (roleNames.length > MAX_ROLES && !_.contains(loadedRoles, roleNames[i])) {
                roleNames.splice(i, 1);
              }
            }
          }

          return roleNames;
        });
    };

    builder.getValueFor = function (key) {
      return _.contains(this.getDataSource().split(','), key);
    };

    builder.getConfigFor = function (key) {
      return { properties: { trueText: key, falseText: key }, formFactor: 'mini' };
    };

    builder.getDomFor = function () {
      return $('<div/>').appendTo(that.$getRolesContainerElement());
    };

    return builder;
  };

  /**
   * Read what roles are selected and join them together into a string.
   *
   * @returns {Promise} promise to be resolved with a comma-separated
   * string containing what roles are selected
   */
  RolesEditor.prototype.doRead = function () {
    var builder = this.getBuilder();
    return Promise.join(builder.getKeys(), builder.readAll())
      .spread(function (roleNames, checks) {
        return _.filter(roleNames, function (roleName, i) { return checks[i]; })
          .join(',');
      });
  };

  /**
   * Remove "RolesEditor" class.
   *
   * @returns {Promise}
   */
  RolesEditor.prototype.doDestroy = function () {
    this.jq().removeClass('RolesEditor');
    return CompositeEditor.prototype.doDestroy.apply(this, arguments);
  };

  return (RolesEditor);
});
